import logging
from typing import List, Tuple

import pandas as pd
import tiktoken
from openai import AsyncOpenAI
from tiktoken import Encoding

from autorag.nodes.generator.base import BaseGenerator
from autorag.utils.util import (
	get_event_loop,
	process_batch,
	pop_params,
	result_to_dataframe,
)

logger = logging.getLogger("AutoRAG")

MAX_TOKEN_DICT = {  # model name : token limit
	"gpt-4.5-preview": 128_000,
	"gpt-4.5-preview-2025-02-27": 128_000,
	"o1": 200_000,
	"o1-preview": 128_000,
	"o1-preview-2024-09-12": 128_000,
	"o1-mini": 128_000,
	"o1-mini-2024-09-12": 128_000,
	"o3-mini": 200_000,
	"gpt-4o-mini": 128_000,
	"gpt-4o-mini-2024-07-18": 128_000,
	"gpt-4o": 128_000,
	"gpt-4o-2024-08-06": 128_000,
	"gpt-4o-2024-05-13": 128_000,
	"chatgpt-4o-latest": 128_000,
	"gpt-4-turbo": 128_000,
	"gpt-4-turbo-2024-04-09": 128_000,
	"gpt-4-turbo-preview": 128_000,
	"gpt-4-0125-preview": 128_000,
	"gpt-4-1106-preview": 128_000,
	"gpt-4-vision-preview": 128_000,
	"gpt-4-1106-vision-preview": 128_000,
	"gpt-4": 8_192,
	"gpt-4-0613": 8_192,
	"gpt-4-32k": 32_768,
	"gpt-4-32k-0613": 32_768,
	"gpt-3.5-turbo-0125": 16_385,
	"gpt-3.5-turbo": 16_385,
	"gpt-3.5-turbo-1106": 16_385,
	"gpt-3.5-turbo-instruct": 4_096,
	"gpt-3.5-turbo-16k": 16_385,
	"gpt-3.5-turbo-0613": 4_096,
	"gpt-3.5-turbo-16k-0613": 16_385,
}


class OpenAILLM(BaseGenerator):
	def __init__(self, project_dir, llm: str, batch: int = 16, *args, **kwargs):
		super().__init__(project_dir, llm, *args, **kwargs)
		assert batch > 0, "batch size must be greater than 0."
		self.batch = batch

		client_init_params = pop_params(AsyncOpenAI.__init__, kwargs)
		self.client = AsyncOpenAI(**client_init_params)

		if self.llm.startswith("gpt-4.5"):
			self.tokenizer = tiktoken.get_encoding("o200k_base")
		else:
			self.tokenizer = tiktoken.encoding_for_model(self.llm)

		self.max_token_size = (
			MAX_TOKEN_DICT.get(self.llm) - 7
		)  # because of chat token usage
		if self.max_token_size is None:
			raise ValueError(
				f"Model {self.llm} does not supported. "
				f"Please select the model between {list(MAX_TOKEN_DICT.keys())}"
			)

	@result_to_dataframe(["generated_texts", "generated_tokens", "generated_log_probs"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		prompts = self.cast_to_run(previous_result)
		return self._pure(prompts, **kwargs)

	def _pure(
		self,
		prompts: List[str],
		truncate: bool = True,
		**kwargs,
	) -> Tuple[List[str], List[List[int]], List[List[float]]]:
		"""
		OpenAI generator module.
		Uses an official openai library for generating answer from the given prompt.
		It returns real token ids and log probs, so you must use this for using token ids and log probs.

		:param prompts: A list of prompts.
		:param llm: A model name for openai.
		    Default is gpt-3.5-turbo.
		:param batch: Batch size for openai api call.
		    If you get API limit errors, you should lower the batch size.
		    Default is 16.
		:param truncate: Whether to truncate the input prompt.
		    Default is True.
		:param api_key: OpenAI API key. You can set this by passing env variable `OPENAI_API_KEY`
		:param kwargs: The optional parameter for openai api call `openai.chat.completion`
		    See https://platform.openai.com/docs/api-reference/chat/create for more details.
		:return: A tuple of three elements.
		    The first element is a list of generated text.
		    The second element is a list of generated text's token ids.
		    The third element is a list of generated text's log probs.
		"""
		if kwargs.get("logprobs") is not None:
			kwargs.pop("logprobs")
			logger.warning(
				"parameter logprob does not effective. It always set to True."
			)
		if kwargs.get("n") is not None:
			kwargs.pop("n")
			logger.warning("parameter n does not effective. It always set to 1.")

		# TODO: fix this after updating tiktoken for the gpt-4.5 model. It is not yet supported yet.
		if truncate:
			prompts = list(
				map(
					lambda prompt: truncate_by_token(
						prompt, self.tokenizer, self.max_token_size
					),
					prompts,
				)
			)

		openai_chat_params = pop_params(self.client.chat.completions.create, kwargs)
		loop = get_event_loop()
		if self.llm.startswith("o1") or self.llm.startswith("o3"):
			tasks = [
				self.get_result_o1(prompt, **openai_chat_params) for prompt in prompts
			]
		else:
			tasks = [
				self.get_result(prompt, **openai_chat_params) for prompt in prompts
			]
		result = loop.run_until_complete(process_batch(tasks, self.batch))
		answer_result = list(map(lambda x: x[0], result))
		token_result = list(map(lambda x: x[1], result))
		logprob_result = list(map(lambda x: x[2], result))
		return answer_result, token_result, logprob_result

	def structured_output(self, prompts: List[str], output_cls, **kwargs):
		supported_models = [
			"gpt-4o-mini-2024-07-18",
			"gpt-4o-2024-08-06",
		]
		if self.llm not in supported_models:
			raise ValueError(
				f"{self.llm} is not a valid model name for structured output. "
				f"Please select the model between {supported_models}"
			)

		if kwargs.get("logprobs") is not None:
			kwargs.pop("logprobs")
			logger.warning(
				"parameter logprob does not effective. It always set to False."
			)
		if kwargs.get("n") is not None:
			kwargs.pop("n")
			logger.warning("parameter n does not effective. It always set to 1.")

		# TODO: fix this after updating tiktoken for the gpt-4.5 model. It is not yet supported yet.
		prompts = list(
			map(
				lambda prompt: truncate_by_token(
					prompt, self.tokenizer, self.max_token_size
				),
				prompts,
			)
		)

		openai_chat_params = pop_params(self.client.beta.chat.completions.parse, kwargs)
		loop = get_event_loop()
		tasks = [
			self.get_structured_result(prompt, output_cls, **openai_chat_params)
			for prompt in prompts
		]
		result = loop.run_until_complete(process_batch(tasks, self.batch))
		return result

	async def astream(self, prompt: str, **kwargs):
		# TODO: gpt-4.5-preview does not support logprobs. It should be fixed after the openai update.
		if kwargs.get("logprobs") is not None:
			kwargs.pop("logprobs")
			logger.warning(
				"parameter logprob does not effective. It always set to False."
			)
		if kwargs.get("n") is not None:
			kwargs.pop("n")
			logger.warning("parameter n does not effective. It always set to 1.")

		prompt = truncate_by_token(prompt, self.tokenizer, self.max_token_size)

		openai_chat_params = pop_params(self.client.chat.completions.create, kwargs)

		stream = await self.client.chat.completions.create(
			model=self.llm,
			messages=[
				{"role": "user", "content": prompt},
			],
			logprobs=False,
			n=1,
			stream=True,
			**openai_chat_params,
		)
		result = ""
		async for chunk in stream:
			if chunk.choices[0].delta.content is not None:
				result += chunk.choices[0].delta.content
				yield result

	def stream(self, prompt: str, **kwargs):
		raise NotImplementedError("stream method is not implemented yet.")

	async def get_structured_result(self, prompt: str, output_cls, **kwargs):
		logprobs = True
		if self.llm.startswith("gpt-4.5"):
			logprobs = False
		response = await self.client.beta.chat.completions.parse(
			model=self.llm,
			messages=[
				{"role": "user", "content": prompt},
			],
			response_format=output_cls,
			logprobs=logprobs,
			n=1,
			**kwargs,
		)
		return response.choices[0].message.parsed

	async def get_result(self, prompt: str, **kwargs):
		# TODO: gpt-4.5-preview does not support logprobs. It should be fixed after the openai update.
		logprobs = True
		if self.llm.startswith("gpt-4.5"):
			logprobs = False
		response = await self.client.chat.completions.create(
			model=self.llm,
			messages=[
				{"role": "user", "content": prompt},
			],
			logprobs=logprobs,
			n=1,
			**kwargs,
		)
		choice = response.choices[0]
		answer = choice.message.content
		# TODO: gpt-4.5-preview does not support logprobs. It should be fixed after the openai update.
		if self.llm.startswith("gpt-4.5"):
			tokens = self.tokenizer.encode(answer, allowed_special="all")
			logprobs = [0.5] * len(tokens)
			logger.warning("gpt-4.5-preview does not support logprobs yet.")
		else:
			logprobs = list(map(lambda x: x.logprob, choice.logprobs.content))
			tokens = list(
				map(
					lambda x: self.tokenizer.encode(x.token, allowed_special="all")[0],
					choice.logprobs.content,
				)
			)
			assert len(tokens) == len(
				logprobs
			), "tokens and logprobs size is different."
		return answer, tokens, logprobs

	async def get_result_o1(self, prompt: str, **kwargs):
		assert self.llm.startswith("o1") or self.llm.startswith(
			"o3"
		), "This function only supports o1 or o3 model."
		# The default temperature for the o1 model is 1. 1 is only supported.
		# See https://platform.openai.com/docs/guides/reasoning about beta limitation of o1 models.
		kwargs["temperature"] = 1
		kwargs["top_p"] = 1
		kwargs["presence_penalty"] = 0
		kwargs["frequency_penalty"] = 0
		response = await self.client.chat.completions.create(
			model=self.llm,
			messages=[
				{"role": "user", "content": prompt},
			],
			logprobs=False,
			n=1,
			**kwargs,
		)
		answer = response.choices[0].message.content
		tokens = self.tokenizer.encode(answer, allowed_special="all")
		pseudo_log_probs = [0.5] * len(tokens)
		return answer, tokens, pseudo_log_probs


def truncate_by_token(prompt: str, tokenizer: Encoding, max_token_size: int):
	tokens = tokenizer.encode(prompt, allowed_special="all")
	return tokenizer.decode(tokens[:max_token_size])



import logging
import os
import pathlib
import uuid
from copy import deepcopy
from typing import Optional, Dict, List

import pandas as pd
import yaml

from autorag.support import get_support_modules
from autorag.utils.util import load_summary_file, load_yaml_config

logger = logging.getLogger("AutoRAG")


def extract_node_line_names(config_dict: Dict) -> List[str]:
	"""
	Extract node line names with the given config dictionary order.

	:param config_dict: The YAML configuration dict for the pipeline.
	    You can load this to access trail_folder/config.yaml.
	:return: The list of node line names.
	    It is the order of the node line names in the pipeline.
	"""
	return [node_line["node_line_name"] for node_line in config_dict["node_lines"]]


def extract_node_strategy(config_dict: Dict) -> Dict:
	"""
	Extract node strategies with the given config dictionary.
	The return value is a dictionary of the node type and its strategy.

	:param config_dict: The YAML configuration dict for the pipeline.
	    You can load this to access trail_folder/config.yaml.
	:return: Key is node_type and value is strategy dict.
	"""
	return {
		node["node_type"]: node.get("strategy", {})
		for node_line in config_dict["node_lines"]
		for node in node_line["nodes"]
	}


def summary_df_to_yaml(summary_df: pd.DataFrame, config_dict: Dict) -> Dict:
	"""
	Convert trial summary dataframe to config yaml file.

	:param summary_df: The trial summary dataframe of the evaluated trial.
	:param config_dict: The yaml configuration dict for the pipeline.
	    You can load this to access trail_folder/config.yaml.
	:return: Dictionary of config yaml file.
	    You can save this dictionary to yaml file.
	"""

	# summary_df columns : 'node_line_name', 'node_type', 'best_module_filename',
	#                      'best_module_name', 'best_module_params', 'best_execution_time'
	node_line_names = extract_node_line_names(config_dict)
	node_strategies = extract_node_strategy(config_dict)
	strategy_df = pd.DataFrame(
		{
			"node_type": list(node_strategies.keys()),
			"strategy": list(node_strategies.values()),
		}
	)
	summary_df = summary_df.merge(strategy_df, on="node_type", how="left")
	summary_df["categorical_node_line_name"] = pd.Categorical(
		summary_df["node_line_name"], categories=node_line_names, ordered=True
	)
	summary_df = summary_df.sort_values(by="categorical_node_line_name")
	grouped = summary_df.groupby("categorical_node_line_name", observed=False)

	node_lines = [
		{
			"node_line_name": node_line_name,
			"nodes": [
				{
					"node_type": row["node_type"],
					"strategy": row["strategy"],
					"modules": [
						{
							"module_type": row["best_module_name"],
							**row["best_module_params"],
						}
					],
				}
				for _, row in node_line.iterrows()
			],
		}
		for node_line_name, node_line in grouped
	]
	return {"node_lines": node_lines}


def extract_best_config(trial_path: str, output_path: Optional[str] = None) -> Dict:
	"""
	Extract the optimal pipeline from the evaluated trial.

	:param trial_path: The path to the trial directory that you want to extract the pipeline from.
	    Must already be evaluated.
	:param output_path: Output path that pipeline yaml file will be saved.
	    Must be .yaml or .yml file.
	    If None, it does not save YAML file and just returns dict values.
	    Default is None.
	:return: The dictionary of the extracted pipeline.
	"""
	summary_path = os.path.join(trial_path, "summary.csv")
	if not os.path.exists(summary_path):
		raise ValueError(f"summary.csv does not exist in {trial_path}.")
	trial_summary_df = load_summary_file(
		summary_path, dict_columns=["best_module_params"]
	)
	config_yaml_path = os.path.join(trial_path, "config.yaml")
	with open(config_yaml_path, "r") as f:
		config_dict = yaml.safe_load(f)
	yaml_dict = summary_df_to_yaml(trial_summary_df, config_dict)
	yaml_dict["vectordb"] = extract_vectordb_config(trial_path)
	if output_path is not None:
		with open(output_path, "w") as f:
			yaml.safe_dump(yaml_dict, f)
	return yaml_dict


def extract_vectordb_config(trial_path: str) -> List[Dict]:
	# get vectordb.yaml file
	project_dir = pathlib.PurePath(os.path.realpath(trial_path)).parent
	vectordb_config_path = os.path.join(project_dir, "resources", "vectordb.yaml")
	if not os.path.exists(vectordb_config_path):
		raise ValueError(f"vectordb.yaml does not exist in {vectordb_config_path}.")
	with open(vectordb_config_path, "r") as f:
		vectordb_dict = yaml.safe_load(f)
	result = vectordb_dict.get("vectordb", [])
	if len(result) != 0:
		return result
	# return default setting of chroma
	return [
		{
			"name": "default",
			"db_type": "chroma",
			"client_type": "persistent",
			"embedding_model": "openai",
			"collection_name": "openai",
			"path": os.path.join(project_dir, "resources", "chroma"),
		}
	]


class BaseRunner:
	def __init__(self, config: Dict, project_dir: Optional[str] = None):
		self.config = config
		project_dir = os.getcwd() if project_dir is None else project_dir
		os.environ["PROJECT_DIR"] = project_dir

		# init modules
		node_lines = deepcopy(self.config["node_lines"])
		self.module_instances = []
		self.module_params = []
		for node_line in node_lines:
			for node in node_line["nodes"]:
				if len(node["modules"]) != 1:
					raise ValueError(
						"The number of modules in a node must be 1 for using runner."
						"Please use extract_best_config method for extracting yaml file from evaluated trial."
					)
				module = node["modules"][0]
				module_type = module.pop("module_type")
				module_params = module
				module_instance = get_support_modules(module_type)(
					project_dir=project_dir,
					**module_params,
				)
				self.module_instances.append(module_instance)
				self.module_params.append(module_params)

	@classmethod
	def from_yaml(cls, yaml_path: str, project_dir: Optional[str] = None):
		"""
		Load Runner from the YAML file.
		Must be extracted YAML file from the evaluated trial using the extract_best_config method.

		:param yaml_path: The path of the YAML file.
		:param project_dir: The path of the project directory.
			Default is the current directory.
		:return: Initialized Runner.
		"""
		config = load_yaml_config(yaml_path)
		return cls(config, project_dir=project_dir)

	@classmethod
	def from_trial_folder(cls, trial_path: str):
		"""
		Load Runner from the evaluated trial folder.
		Must already be evaluated using Evaluator class.
		It sets the project_dir as the parent directory of the trial folder.

		:param trial_path: The path of the trial folder.
		:return: Initialized Runner.
		"""
		config = extract_best_config(trial_path)
		return cls(config, project_dir=os.path.dirname(trial_path))


class Runner(BaseRunner):
	def run(self, query: str, result_column: str = "generated_texts"):
		"""
		Run the pipeline with query.
		The loaded pipeline must start with a single query,
		so the first module of the pipeline must be `query_expansion` or `retrieval` module.

		:param query: The query of the user.
		:param result_column: The result column name for the answer.
		    Default is `generated_texts`, which is the output of the `generation` module.
		:return: The result of the pipeline.
		"""
		previous_result = pd.DataFrame(
			{
				"qid": str(uuid.uuid4()),
				"query": [query],
				"retrieval_gt": [[]],
				"generation_gt": [""],
			}
		)  # pseudo qa data for execution
		for module_instance, module_param in zip(
			self.module_instances, self.module_params
		):
			new_result = module_instance.pure(
				previous_result=previous_result, **module_param
			)
			duplicated_columns = previous_result.columns.intersection(
				new_result.columns
			)
			drop_previous_result = previous_result.drop(columns=duplicated_columns)
			previous_result = pd.concat([drop_previous_result, new_result], axis=1)

		return previous_result[result_column].tolist()[0]



from .cohere import CohereReranker
from .colbert import ColbertReranker
from .flag_embedding import FlagEmbeddingReranker
from .flag_embedding_llm import FlagEmbeddingLLMReranker
from .jina import JinaReranker
from .koreranker import KoReranker
from .monot5 import MonoT5
from .pass_reranker import PassReranker
from .rankgpt import RankGPT
from .sentence_transformer import SentenceTransformerReranker
from .time_reranker import TimeReranker
from .upr import Upr
from .openvino import OpenVINOReranker
from .voyageai import VoyageAIReranker
from .mixedbreadai import MixedbreadAIReranker
from .flashrank import FlashRankReranker



from llama_index.core.base.llms.types import ChatMessage, MessageRole

FILTER_PROMPT = {
	"dontknow_filter": {
		"en": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""The following sentence is an answer about a question. You have to decide the answer implies 'I don't know'.
If the answer implies 'I don't know', return True. If not, return False.""",
			),
		],
		"ko": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""다음 문장은 어떠한 질문에 대한 대답입니다. 해당 문장이 질문에 대해서 '모른다고' 답한 것인지 판단하십시오.
만약 해당 문장이 '모른다고' 답한 것이라면, True를 반환하세요. 그렇지 않다면 False를 반환하세요.""",
			)
		],
		"ja": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""次の文章はある質問に対する答えです。 該当文章が質問に対して「知らない」と答えたのか判断します。
もし、その文章が「知らない」と答えたのであれば、Trueを返します。 そうでなければFalseを返します。""",
			)
		],
	},
	"passage_dependency": {
		"en": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""You are a classifier that recognize 'passage dependent' questions.
The 'passage dependent' is the question that the answer will be change depending on what passage you choose.
For example) 'What is the highest score according to the table?'
This sentence is the passage dependent question because the answer will be different depending on the table.

In contrast, the following sentence is not passage dependant.
'What is the highest score of the KBO baseball history in one game?'
'What is the capital of France?'
These sentences will have the same answer regardless of the passage.

Please return True if the input question is passage dependent. Else return False.""",
			)
		],
		"ko": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""당신은 '단락 의존' 질문을 인식하는 분류기입니다.
'단락 의존'이란 어떤 단락이 선택 되는지 따라 답이 달라지는 질문을 의미합니다.
예를 들어, '주어진 표에 따르면 가장 높은 점수는 무엇인가요?'라는 질문은 단락 의존 질문입니다. 왜냐하면 표가 어떤 것인지에 따라 그 답이 달라지기 때문입니다.

반면에, 다음 문장들은 단락 의존적이지 않습니다.
'KBO 야구 역사상 한 경기에서 가장 높은 점수는 무엇인가요?' 또는 '프랑스의 수도는 무엇인가요?'
이러한 문장은 단락에 관계 없이 동일한 답을 가집니다.

입력된 질문이 단락 의존적이라면 True를 반환하고, 그렇지 않으면 False를 반환하세요.""",
			)
		],
		"ja": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""あなたは「段落依存」の質問を認識する分類器です。
「段落依存」とは、どの段落が選択されるかによって答えが変わる質問を意味します。
たとえば、「与えられた表によると、最も高い点数は何ですか？」という質問は、段落依存の質問です。 なぜなら、表がどんなものかによってその答えが変わるからです。

一方、次の文章は段落依存的ではありません。
KBO野球史上1試合で最も高い点数は何ですか?またはフランスの首都は何ですか?'
このような文章は段落に関係なく同じ答えを持ちます。

入力された質問が段落依存的である場合はTrueを返し、そうでない場合はFalseを返します。""",
			)
		],
	},
}



import os
import pathlib
from typing import Dict, List, Optional

import pandas as pd

from autorag.schema import Node
from autorag.utils.util import load_summary_file


def make_node_lines(node_line_dict: Dict) -> List[Node]:
	"""
	This method makes a list of nodes from node line dictionary.
	:param node_line_dict: Node_line_dict loaded from yaml file, or get from user input.
	:return: List of Nodes inside this node line.
	"""
	nodes = node_line_dict.get("nodes")
	if nodes is None:
		raise ValueError("Node line must have 'nodes' key.")
	node_objects = list(map(lambda x: Node.from_dict(x), nodes))
	return node_objects


def run_node_line(
	nodes: List[Node],
	node_line_dir: str,
	previous_result: Optional[pd.DataFrame] = None,
):
	"""
	Run the whole node line by running each node.

	:param nodes: A list of nodes.
	:param node_line_dir: This node line's directory.
	:param previous_result: A result of the previous node line.
	    If None, it loads qa data from data/qa.parquet.
	:return: The final result of the node line.
	"""
	if previous_result is None:
		project_dir = pathlib.PurePath(node_line_dir).parent.parent
		qa_path = os.path.join(project_dir, "data", "qa.parquet")
		if not os.path.exists(qa_path):
			raise ValueError(f"qa.parquet does not exist in {qa_path}.")
		previous_result = pd.read_parquet(qa_path, engine="pyarrow")

	summary_lst = []
	for node in nodes:
		previous_result = node.run(previous_result, node_line_dir)
		node_summary_df = load_summary_file(
			os.path.join(node_line_dir, node.node_type, "summary.csv")
		)
		best_node_row = node_summary_df.loc[node_summary_df["is_best"]]
		summary_lst.append(
			{
				"node_type": node.node_type,
				"best_module_filename": best_node_row["filename"].values[0],
				"best_module_name": best_node_row["module_name"].values[0],
				"best_module_params": best_node_row["module_params"].values[0],
				"best_execution_time": best_node_row["execution_time"].values[0],
			}
		)

	pd.DataFrame(summary_lst).to_csv(
		os.path.join(node_line_dir, "summary.csv"), index=False
	)
	return previous_result



from typing import List, Tuple

import pandas as pd

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.nodes.passagereranker.flag_embedding import flag_embedding_run_model
from autorag.utils.util import (
	flatten_apply,
	sort_by_scores,
	select_top_k,
	pop_params,
	result_to_dataframe,
	empty_cuda_cache,
)


class FlagEmbeddingLLMReranker(BasePassageReranker):
	def __init__(
		self,
		project_dir,
		model_name: str = "BAAI/bge-reranker-v2-gemma",
		*args,
		**kwargs,
	):
		"""
		Initialize the FlagEmbeddingReranker module.

		:param project_dir: The project directory.
		:param model_name: The name of the BAAI Reranker LLM-based-model name.
		Default is "BAAI/bge-reranker-v2-gemma"
		:param kwargs: Extra parameter for FlagEmbedding.FlagReranker
		"""
		super().__init__(project_dir)
		try:
			from FlagEmbedding import FlagLLMReranker
		except ImportError:
			raise ImportError(
				"FlagEmbeddingLLMReranker requires the 'FlagEmbedding' package to be installed."
			)
		model_params = pop_params(FlagLLMReranker.__init__, kwargs)
		model_params.pop("model_name_or_path", None)
		self.model = FlagLLMReranker(model_name_or_path=model_name, **model_params)

	def __del__(self):
		del self.model
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		batch = kwargs.pop("batch", 64)
		return self._pure(queries, contents, ids, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents based on their relevance to a query using BAAI LLM-based-Reranker model.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param batch: The number of queries to be processed in a batch
			Default is 64.

		:return: tuple of lists containing the reranked contents, ids, and scores
		"""

		nested_list = [
			list(map(lambda x: [query, x], content_list))
			for query, content_list in zip(queries, contents_list)
		]
		rerank_scores = flatten_apply(
			flag_embedding_run_model, nested_list, model=self.model, batch_size=batch
		)

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": rerank_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)



import mimetypes
import os
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Tuple, Callable

import pandas as pd
import yaml
from langchain_core.documents import Document
from llama_index.core.schema import NodeRelationship

from autorag.schema import Module
from autorag.utils.util import make_combinations, explode


def get_file_metadata(file_path: str) -> Dict:
	"""Get some handy metadate from filesystem.

	Args:
	    file_path: str: file path in str
	"""
	return {
		"file_path": file_path,
		"file_name": os.path.basename(file_path),
		"file_type": mimetypes.guess_type(file_path)[0],
		"file_size": os.path.getsize(file_path),
		"creation_datetime": datetime.fromtimestamp(
			Path(file_path).stat().st_ctime
		).strftime("%Y-%m-%d"),
		"last_modified_datetime": datetime.fromtimestamp(
			Path(file_path).stat().st_mtime
		).strftime("%Y-%m-%d"),
		"last_accessed_datetime": datetime.fromtimestamp(
			Path(file_path).stat().st_atime
		).strftime("%Y-%m-%d"),
	}


def add_essential_metadata(metadata: Dict) -> Dict:
	if "last_modified_datetime" not in metadata:
		metadata["last_modified_datetime"] = datetime.now()
	return metadata


def corpus_df_to_langchain_documents(corpus_df: pd.DataFrame) -> List[Document]:
	page_contents = corpus_df["contents"].tolist()
	ids = corpus_df["doc_id"].tolist()
	metadatas = corpus_df["metadata"].tolist()
	return list(
		map(
			lambda x: Document(page_content=x[0], metadata={"filename": x[1], **x[2]}),
			zip(page_contents, ids, metadatas),
		)
	)


def add_essential_metadata_llama_text_node(metadata: Dict, relationships: Dict) -> Dict:
	if "last_modified_datetime" not in metadata:
		metadata["last_modified_datetime"] = datetime.now()

	if "prev_id" not in metadata:
		if NodeRelationship.PREVIOUS in relationships:
			prev_node = relationships.get(NodeRelationship.PREVIOUS, None)
			if prev_node:
				metadata["prev_id"] = prev_node.node_id

	if "next_id" not in metadata:
		if NodeRelationship.NEXT in relationships:
			next_node = relationships.get(NodeRelationship.NEXT, None)
			if next_node:
				metadata["next_id"] = next_node.node_id
	return metadata


def load_yaml(yaml_path: str):
	if not os.path.exists(yaml_path):
		raise ValueError(f"YAML file {yaml_path} does not exist.")
	with open(yaml_path, "r", encoding="utf-8") as stream:
		try:
			yaml_dict = yaml.safe_load(stream)
		except yaml.YAMLError as exc:
			raise ValueError(f"YAML file {yaml_path} could not be loaded.") from exc
	return yaml_dict["modules"]


def get_param_combinations(modules: List[Dict]) -> Tuple[List[Callable], List[Dict]]:
	module_callable_list, module_params_list = [], []
	for module in modules:
		module_instance = Module.from_dict(module)
		module_params_list.append(module_instance.module_param)
		module_callable_list.append(module_instance.module)

	combinations = list(map(make_combinations, module_params_list))
	module_list, combination_list = explode(module_callable_list, combinations)
	return module_list, combination_list


def get_start_end_idx(original_text: str, search_str: str) -> Tuple[int, int]:
	start_idx = original_text.find(search_str)
	if start_idx == -1:
		return 0, 0
	end_idx = start_idx + len(search_str)
	return start_idx, end_idx - 1



import pandas as pd

from autorag.nodes.queryexpansion.base import BaseQueryExpansion
from autorag.utils import result_to_dataframe


class PassQueryExpansion(BaseQueryExpansion):
	@result_to_dataframe(["queries"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		"""
		Do not perform query expansion.
		Return with the same queries.
		The dimension will be 2-d list, and the column name will be 'queries'.
		"""
		assert (
			"query" in previous_result.columns
		), "previous_result must have query column."
		queries = previous_result["query"].tolist()
		return list(map(lambda x: [x], queries))

	def _pure(self, *args, **kwargs):
		pass



class FaithfulnessTemplate:
	@staticmethod
	def generate_claims(text, lang: str = "en"):
		if lang == "en":
			return f"""Based on the given text, please generate a comprehensive list of FACTUAL claims that can inferred from the provided text.

	Example:
	Example Text:
	"Einstein won the noble prize in 1968 for his discovery of the photoelectric effect."

	Example JSON:
	{{
	    "claims": [
	        "Einstein won the noble prize for his discovery of the photoelectric effect.",
	        "Einstein won the noble prize in 1968."
	    ]
	}}
	===== END OF EXAMPLE ======

	**
	IMPORTANT: Please make sure to only return in JSON format, with the "claims" key as a list of strings. No words or explanation is needed.
	Only include claims that are factual, and the claims you extract should include the full context it was presented in, NOT cherry picked facts.
	You should NOT include any prior knowledge, and take the text at face value when extracting claims.
	**

	Text:
	{text}

	JSON:
	"""
		elif lang == "ko":
			return f"""주어진 텍스트에서 찾을 수 있는 사실적 정보들의 목록을 생성하세요.

예시:
예시 텍스트:
“아인슈타인은 1968년에 광전 효과 발견으로 노벨상을 수상했다.”

예시 JSON:
{{
“claims”: [
“아인슈타인은 광전 효과 발견으로 노벨상을 수상했다.”,
“아인슈타인은 1968년에 노벨상을 수상했다.”
]
}}
===== 예시 끝 ======

**
중요: 오직 JSON 형식으로 “claims” 키가 문자열 목록으로 반환되도록 해야 합니다. 다른 단어나 설명은 필요하지 않습니다.
사실에 기반한 주장만 포함하며, 추출한 주장은 전체 맥락을 유지해야 하며, 부분적으로 선택된 사실을 포함하지 않아야 합니다.
사전 지식은 포함하지 말고, 텍스트에만 기초해 주장들을 추출해야 합니다.
**

텍스트:
{text}

JSON:
"""
		elif lang == "ja":
			return f"""与えられたテキストに基づいて、そこから推測できる事実に基づく主張のリストを生成してください。

例:
例のテキスト:
「アインシュタインは1968年に光電効果の発見でノーベル賞を受賞しました。」

例のJSON:
{{
    "claims": [
        "アインシュタインは光電効果の発見でノーベル賞を受賞しました。",
        "アインシュタインは1968年にノーベル賞を受賞しました。"
    ]
}}
===== 例の終わり ======

**
重要: 必ずJSON形式で"claims"キーが文字列のリストとして返されるようにしてください。説明や余計な言葉は不要です。
事実に基づく主張のみを含め、抽出された主張は提示された文脈全体を含むものでなければなりません。一部の事実のみを抜粋することは避けてください。
事前知識を使用せず、テキストに基づいて主張を抽出してください。
**

テキスト:
{text}

JSON:
"""
		else:
			raise ValueError(f"Language {lang} is not supported.")

	@staticmethod
	def generate_truths(text, lang: str = "en"):
		if lang == "en":
			return f"""Based on the given text, please generate a comprehensive list of FACTUAL, undisputed truths that can inferred from the provided text.

	Example:
	Example Text:
	"Einstein won the noble prize in 1968 for his discovery of the photoelectric effect."

	Example JSON:
	{{
	    "truths": [
	        "Einstein won the noble prize for his discovery of the photoelectric effect.",
	        "Einstein won the noble prize in 1968."
	    ]
	}}
	===== END OF EXAMPLE ======

	**
	IMPORTANT: Please make sure to only return in JSON format, with the "truths" key as a list of strings. No words or explanation is needed.
	Only include truths that are factual.
	**

	Text:
	{text}

	JSON:
	"""
		elif lang == "ko":
			return f"""주어진 텍스트에서 추출할 수 있는 사실적이고 논란이 없는 진실들의 목록을 생성하세요.

예시:
예시 텍스트:
"아인슈타인은 1968년에 광전 효과 발견으로 노벨상을 수상했다."

예시 JSON:
{{
    "truths": [
        "아인슈타인은 광전 효과 발견으로 노벨상을 수상했다.",
        "아인슈타인은 1968년에 노벨상을 수상했다."
    ]
}}
===== 예시 끝 ======

**
중요: 오직 JSON 형식으로 "truths" 키가 문자열 목록으로 반환되도록 해야 합니다. 다른 단어나 설명은 필요하지 않습니다.
사실에 기반한 진실만 포함해야 합니다.
**

텍스트:
{text}

JSON:
"""
		elif lang == "ja":
			return f"""与えられたテキストに基づいて、そこから推測できる事実で議論の余地のない真実のリストを生成してください。

例:
例のテキスト:
「アインシュタインは1968年に光電効果の発見でノーベル賞を受賞しました。」

例のJSON:
{{
    "truths": [
        "アインシュタインは光電効果の発見でノーベル賞を受賞しました。",
        "アインシュタインは1968年にノーベル賞を受賞しました。"
    ]
}}
===== 例の終わり ======

**
重要: 必ずJSON形式で"truths"キーが文字列のリストとして返されるようにしてください。説明や余計な言葉は不要です。
事実に基づく真実のみを含めてください。
**

テキスト:
{text}

JSON:
"""
		else:
			raise ValueError(f"Language {lang} is not supported.")

	@staticmethod
	def generate_verdicts(claims, retrieval_context, lang: str = "en"):
		if lang == "en":
			return f"""Based on the given claims, which is a list of strings, generate a list of JSON objects to indicate whether EACH claim contradicts any facts in the retrieval context. The JSON will have 2 fields: 'verdict' and 'reason'.
	The 'verdict' key should STRICTLY be either 'yes', 'no', or 'idk', which states whether the given claim agrees with the context.
	Provide a 'reason' ONLY if the answer is 'no'.
	The provided claim is drawn from the actual output. Try to provide a correction in the reason using the facts in the retrieval context.

	**
	IMPORTANT: Please make sure to only return in JSON format, with the 'verdicts' key as a list of JSON objects.
	Example retrieval contexts: "Einstein won the Nobel Prize for his discovery of the photoelectric effect. Einstein won the Nobel Prize in 1968. Einstein is a German Scientist."
	Example claims: ["Barack Obama is a caucasian male.", "Zurich is a city in London", "Einstein won the Nobel Prize for the discovery of the photoelectric effect which may have contributed to his fame.", "Einstein won the Nobel Prize in 1969 for his discovery of the photoelectric effect.", "Einstein was a Germen chef."]

	Example:
	{{
	    "verdicts": [
	        {{
	            "verdict": "idk"
	        }},
	        {{
	            "verdict": "idk"
	        }},
	        {{
	            "verdict": "yes"
	        }},
	        {{
	            "verdict": "no",
	            "reason": "The actual output claims Einstein won the Nobel Prize in 1969, which is untrue as the retrieval context states it is 1968 instead."
	        }},
	        {{
	            "verdict": "no",
	            "reason": "The actual output claims Einstein is a Germen chef, which is not correct as the retrieval context states he was a German scientist instead."
	        }},
	    ]
	}}
	===== END OF EXAMPLE ======

	The length of 'verdicts' SHOULD BE STRICTLY EQUAL to that of claims.
	You DON'T have to provide a reason if the answer is 'yes' or 'idk'.
	ONLY provide a 'no' answer if the retrieval context DIRECTLY CONTRADICTS the claims. YOU SHOULD NEVER USE YOUR PRIOR KNOWLEDGE IN YOUR JUDGEMENT.
	Claims made using vague, suggestive, speculative language such as 'may have', 'possibility due to', does NOT count as a contradiction.
	Claims that is not backed up due to a lack of information/is not mentioned in the retrieval contexts MUST be answered 'idk', otherwise I WILL DIE.
	**

	Retrieval Contexts:
	{retrieval_context}

	Claims:
	{claims}

	JSON:
	"""
		elif lang == "ko":
			return f"""주어진 주장에 대해, 각 주장이 주어진 문맥의 사실들과 모순되는지를 나타내는 JSON 객체 목록을 생성하세요. JSON은 두 개의 필드인 'verdict'와 'reason'으로 구성됩니다.
'verdict'는 'yes', 'no', 또는 'idk' 중 하나여야 하며, 주어진 주장이 문맥과 일치하는지를 나타냅니다.
'verdict'가 'no'인 경우에만 'reason'을 제공하세요. 'reason'에는 문맥에 따라 주장을 수정하는 내용이 포함되어야 합니다.

**
중요: 오직 JSON 형식으로 'verdicts' 키가 JSON 객체 목록으로 반환되도록 해야 합니다.
예시 문맥: "아인슈타인은 광전 효과 발견으로 노벨상을 수상했다. 아인슈타인은 1968년에 노벨상을 수상했다. 아인슈타인은 독일 과학자이다."
예시 주장: ["버락 오바마는 백인 남성이다.", "취리히는 런던에 있는 도시이다.", "아인슈타인은 광전 효과 발견으로 노벨상을 수상했으며, 이는 그의 명성에 기여했을 것이다.", "아인슈타인은 1969년에 광전 효과 발견으로 노벨상을 수상했다.", "아인슈타인은 독일 요리사였다."]

예시:
{{
    "verdicts": [
        {{
            "verdict": "idk"
        }},
        {{
            "verdict": "idk"
        }},
        {{
            "verdict": "yes"
        }},
        {{
            "verdict": "no",
            "reason": "실제 출력은 아인슈타인이 1969년에 노벨상을 수상했다고 주장하지만, 문맥에서는 1968년이라고 명시되어 있습니다."
        }},
        {{
            "verdict": "no",
            "reason": "실제 출력은 아인슈타인이 독일 요리사라고 주장하지만, 문맥에서는 그가 독일 과학자라고 명시되어 있습니다."
        }},
    ]
}}
===== 예시 끝 ======

'verdicts' 리스트의 길이는 반드시 주장들의 길이와 같아야 합니다.
'yes' 또는 'idk'일 경우 'reason'을 제공할 필요가 없습니다.
검색된 문맥과 직접적으로 모순되는 경우에만 'no' 답변을 제공하세요. 절대로 선험적인 지식을 사용하지 마세요.
'~일 수 있다', '가능성이 있다'와 같은 모호한 표현은 모순으로 간주하지 마세요.
문맥에 대한 정보 부족으로 뒷받침되지 않거나 언급되지 않은 주장은 반드시 'idk'로 답변하세요, 그렇지 않으면 내가 죽습니다.
**

주어진 문맥:
{retrieval_context}

주장:
{claims}

JSON:
"""
		elif lang == "ja":
			return f"""与えられた主張について、それぞれの主張が取得された文脈の事実と矛盾しているかどうかを示すJSONオブジェクトのリストを生成してください。JSONには2つのフィールド、'verdict'と'reason'があります。
'verdict'フィールドは、主張が文脈に一致するかどうかを示すため、厳密に'yes', 'no', 'idk'のいずれかを使用します。
'verdict'が'no'の場合にのみ、'reason'を提供してください。'reason'には、文脈に基づいて主張を修正する内容が含まれている必要があります。

**
重要: 必ずJSON形式で'verdicts'キーがJSONオブジェクトのリストとして返されるようにしてください。
例の文脈:「アインシュタインは光電効果の発見でノーベル賞を受賞しました。アインシュタインは1968年にノーベル賞を受賞しました。アインシュタインはドイツの科学者です。」
例の主張: ["バラク・オバマは白人男性です。", "チューリッヒはロンドンにある都市です。", "アインシュタインは光電効果の発見でノーベル賞を受賞し、これが彼の名声に貢献したかもしれません。", "アインシュタインは1969年に光電効果の発見でノーベル賞を受賞しました。", "アインシュタインはドイツのシェフでした。"]

例のJSON:
{{
    "verdicts": [
        {{
            "verdict": "idk"
        }},
        {{
            "verdict": "idk"
        }},
        {{
            "verdict": "yes"
        }},
        {{
            "verdict": "no",
            "reason": "実際の出力は、アインシュタインが1969年にノーベル賞を受賞したと主張していますが、文脈では1968年と述べられています。"
        }},
        {{
            "verdict": "no",
            "reason": "実際の出力は、アインシュタインがドイツのシェフだと主張していますが、文脈では彼がドイツの科学者であると述べられています。"
        }},
    ]
}}
===== 例の終わり ======

'verdicts'のリストの長さは、主張のリストの長さと必ず等しくなければなりません。
'yes'または'idk'の場合、'reason'を提供する必要はありません。
文脈と直接矛盾する場合にのみ、'no'を提供してください。決して事前知識を使用しないでください。
「〜かもしれない」や「〜の可能性がある」といった曖昧な表現は矛盾とは見なされません。
情報が不足している、または文脈で言及されていない主張には必ず'idk'で答えてください。さもないと私は死んでしまいます。
**

文脈:
{retrieval_context}

主張:
{claims}

JSON:
"""
		else:
			raise ValueError(f"Language {lang} is not supported.")



from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from llama_index.core.base.llms.types import ChatMessage, ChatResponse
from llama_index.core.llms import LLM
from llama_index.core.postprocessor.rankGPT_rerank import RankGPTRerank
from llama_index.core.schema import NodeWithScore, QueryBundle, TextNode
from llama_index.core.utils import print_text
from llama_index.llms.openai import OpenAI

from autorag import generator_models
from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import (
	get_event_loop,
	process_batch,
	pop_params,
	result_to_dataframe,
	empty_cuda_cache,
)


class RankGPT(BasePassageReranker):
	def __init__(
		self, project_dir: str, llm: Optional[Union[str, LLM]] = None, **kwargs
	):
		"""
		Initialize the RankGPT reranker.

		:param project_dir: The project directory
		:param llm: The LLM model to use for RankGPT rerank.
			It is a llama index model.
			Default is the OpenAI model with gpt-4o-mini.
		:param kwargs: The keyword arguments for the LLM model.
		"""
		super().__init__(project_dir)
		if llm is None:
			self.llm = OpenAI(model="gpt-4o-mini")
		else:
			if not isinstance(llm, LLM):
				llm_class = generator_models[llm]
				llm_param = pop_params(llm_class.__init__, kwargs)
				self.llm = llm_class(**llm_param)
			else:
				self.llm = llm

	def __del__(self):
		del self.llm
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, scores, ids = self.cast_to_run(previous_result)
		top_k = kwargs.get("top_k", 1)
		verbose = kwargs.get("verbose", False)
		rankgpt_rerank_prompt = kwargs.get("rankgpt_rerank_prompt", None)
		batch = kwargs.get("batch", 16)
		return self._pure(
			queries=queries,
			contents_list=contents,
			scores_list=scores,
			ids_list=ids,
			top_k=top_k,
			verbose=verbose,
			rankgpt_rerank_prompt=rankgpt_rerank_prompt,
			batch=batch,
		)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		scores_list: List[List[float]],
		ids_list: List[List[str]],
		top_k: int,
		verbose: bool = False,
		rankgpt_rerank_prompt: Optional[str] = None,
		batch: int = 16,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank given context paragraphs using RankGPT.
		Return pseudo scores, since the actual scores are not available on RankGPT.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param scores_list: The list of lists of scores retrieved from the initial ranking
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param verbose: Whether to print intermediate steps.
		:param rankgpt_rerank_prompt: The prompt template for RankGPT rerank.
			Default is RankGPT's default prompt.
		:param batch: The number of queries to be processed in a batch.
		:return: Tuple of lists containing the reranked contents, ids, and scores
		"""
		query_bundles = list(map(lambda query: QueryBundle(query_str=query), queries))
		nodes_list = [
			list(
				map(
					lambda x: NodeWithScore(node=TextNode(text=x[0]), score=x[1]),
					zip(content_list, score_list),
				)
			)
			for content_list, score_list in zip(contents_list, scores_list)
		]

		reranker = AsyncRankGPTRerank(
			top_n=top_k,
			llm=self.llm,
			verbose=verbose,
			rankgpt_rerank_prompt=rankgpt_rerank_prompt,
		)

		tasks = [
			reranker.async_postprocess_nodes(nodes, query, ids)
			for nodes, query, ids in zip(nodes_list, query_bundles, ids_list)
		]
		loop = get_event_loop()
		rerank_result = loop.run_until_complete(process_batch(tasks, batch_size=batch))
		content_result = [
			list(map(lambda x: x.node.text, res[0])) for res in rerank_result
		]
		score_result = [
			np.linspace(1.0, 0.0, len(res[0])).tolist() for res in rerank_result
		]
		id_result = [res[1] for res in rerank_result]

		del reranker

		return content_result, id_result, score_result


class AsyncRankGPTRerank(RankGPTRerank):
	async def async_run_llm(self, messages: Sequence[ChatMessage]) -> ChatResponse:
		return await self.llm.achat(messages)

	async def async_postprocess_nodes(
		self,
		nodes: List[NodeWithScore],
		query_bundle: QueryBundle,
		ids: Optional[List[str]] = None,
	) -> Tuple[List[NodeWithScore], List[str]]:
		if ids is None:
			ids = [str(i) for i in range(len(nodes))]

		items = {
			"query": query_bundle.query_str,
			"hits": [{"content": node.get_content()} for node in nodes],
		}

		messages = self.create_permutation_instruction(item=items)
		permutation = await self.async_run_llm(messages=messages)
		if permutation.message is not None and permutation.message.content is not None:
			rerank_ranks = self._receive_permutation(
				items, str(permutation.message.content)
			)
			if self.verbose:
				print_text(f"After Reranking, new rank list for nodes: {rerank_ranks}")

			initial_results: List[NodeWithScore] = []
			id_results = []

			for idx in rerank_ranks:
				initial_results.append(
					NodeWithScore(node=nodes[idx].node, score=nodes[idx].score)
				)
				id_results.append(ids[idx])
			return initial_results[: self.top_n], id_results[: self.top_n]
		else:
			return nodes[: self.top_n], ids[: self.top_n]



import itertools
import logging
import os
import tempfile

import pandas as pd

from autorag.evaluator import Evaluator
from autorag.utils import (
	cast_qa_dataset,
	cast_corpus_dataset,
	validate_qa_from_corpus_dataset,
)

logger = logging.getLogger("AutoRAG")


class Validator:
	def __init__(self, qa_data_path: str, corpus_data_path: str):
		"""
		Initialize a Validator object.

		:param qa_data_path: The path to the QA dataset.
		    Must be parquet file.
		:param corpus_data_path: The path to the corpus dataset.
		    Must be parquet file.
		"""
		# validate data paths
		if not os.path.exists(qa_data_path):
			raise ValueError(f"QA data path {qa_data_path} does not exist.")
		if not os.path.exists(corpus_data_path):
			raise ValueError(f"Corpus data path {corpus_data_path} does not exist.")
		if not qa_data_path.endswith(".parquet"):
			raise ValueError(f"QA data path {qa_data_path} is not a parquet file.")
		if not corpus_data_path.endswith(".parquet"):
			raise ValueError(
				f"Corpus data path {corpus_data_path} is not a parquet file."
			)
		self.qa_data = pd.read_parquet(qa_data_path, engine="pyarrow")
		self.corpus_data = pd.read_parquet(corpus_data_path, engine="pyarrow")
		self.qa_data = cast_qa_dataset(self.qa_data)
		self.corpus_data = cast_corpus_dataset(self.corpus_data)

	def validate(self, yaml_path: str, qa_cnt: int = 5, random_state: int = 42):
		# Determine the sample size and log a warning if qa_cnt is larger than available records
		available_records = len(self.qa_data)
		safe_sample_size = min(qa_cnt, available_records)  # 먼저 safe_sample_size 계산

		if safe_sample_size < qa_cnt:
			logger.warning(
				f"Minimal Requested sample size ({qa_cnt}) is larger than available records ({available_records}). "
				f"Sampling will be limited to {safe_sample_size} records. "
			)

		# safe sample QA data
		sample_qa_df = self.qa_data.sample(
			n=safe_sample_size, random_state=random_state
		)
		sample_qa_df.reset_index(drop=True, inplace=True)

		# get doc_id
		temp_qa_df = sample_qa_df.copy(deep=True)
		flatten_retrieval_gts = (
			temp_qa_df["retrieval_gt"]
			.apply(lambda x: list(itertools.chain.from_iterable(x)))
			.tolist()
		)
		target_doc_ids = list(itertools.chain.from_iterable(flatten_retrieval_gts))

		# make sample corpus data
		sample_corpus_df = self.corpus_data.loc[
			self.corpus_data["doc_id"].isin(target_doc_ids)
		]
		sample_corpus_df.reset_index(drop=True, inplace=True)

		validate_qa_from_corpus_dataset(sample_qa_df, sample_corpus_df)

		# start Evaluate at temp project directory
		with (
			tempfile.NamedTemporaryFile(suffix=".parquet", delete=False) as qa_path,
			tempfile.NamedTemporaryFile(suffix=".parquet", delete=False) as corpus_path,
			tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_project_dir,
		):
			sample_qa_df.to_parquet(qa_path.name, index=False)
			sample_corpus_df.to_parquet(corpus_path.name, index=False)

			evaluator = Evaluator(
				qa_data_path=qa_path.name,
				corpus_data_path=corpus_path.name,
				project_dir=temp_project_dir,
			)
			evaluator.start_trial(yaml_path, skip_validation=True)
			qa_path.close()
			corpus_path.close()
			os.unlink(qa_path.name)
			os.unlink(corpus_path.name)

		logger.info("Validation complete.")



import logging
from typing import Any, Dict, List, Tuple, Optional, Union

from pymilvus import (
	DataType,
	FieldSchema,
	CollectionSchema,
	connections,
	Collection,
	MilvusException,
)
from pymilvus.orm import utility

from autorag.utils.util import apply_recursive
from autorag.vectordb import BaseVectorStore


logger = logging.getLogger("AutoRAG")


class Milvus(BaseVectorStore):
	def __init__(
		self,
		embedding_model: Union[str, List[dict]],
		collection_name: str,
		embedding_batch: int = 100,
		similarity_metric: str = "cosine",
		index_type: str = "IVF_FLAT",
		uri: str = "http://localhost:19530",
		db_name: str = "",
		token: str = "",
		user: str = "",
		password: str = "",
		timeout: Optional[float] = None,
		params: Dict[str, Any] = {},
	):
		super().__init__(embedding_model, similarity_metric, embedding_batch)

		# Connect to Milvus server
		connections.connect(
			"default",
			uri=uri,
			token=token,
			db_name=db_name,
			user=user,
			password=password,
		)
		self.collection_name = collection_name
		self.timeout = timeout
		self.params = params
		self.index_type = index_type

		# Set Collection
		if not utility.has_collection(collection_name, timeout=timeout):
			# Get the dimension of the embeddings
			test_embedding_result: List[float] = self.embedding.get_query_embedding(
				"test"
			)
			dimension = len(test_embedding_result)

			pk = FieldSchema(
				name="id",
				dtype=DataType.VARCHAR,
				max_length=128,
				is_primary=True,
				auto_id=False,
			)
			field = FieldSchema(
				name="vector", dtype=DataType.FLOAT_VECTOR, dim=dimension
			)
			schema = CollectionSchema(fields=[pk, field])

			self.collection = Collection(name=self.collection_name, schema=schema)
			index_params = {
				"metric_type": self.similarity_metric.upper(),
				"index_type": self.index_type.upper(),
				"params": self.params,
			}
			self.collection.create_index(
				field_name="vector", index_params=index_params, timeout=self.timeout
			)
		else:
			self.collection = Collection(name=self.collection_name)

	async def add(self, ids: List[str], texts: List[str]):
		texts = self.truncated_inputs(texts)
		text_embeddings: List[
			List[float]
		] = await self.embedding.aget_text_embedding_batch(texts)
		self.add_embedding(ids, text_embeddings)

	def add_embedding(self, ids: List[str], embeddings: List[List[float]]):
		data = list(
			map(lambda _id, vector: {"id": _id, "vector": vector}, ids, embeddings)
		)

		# Insert data into the collection
		res = self.collection.insert(data=data, timeout=self.timeout)
		assert (
			res.insert_count == len(ids)
		), f"Insertion failed. Try to insert {len(ids)} but only {res['insert_count']} inserted."

		self.collection.flush(timeout=self.timeout)

	async def query(
		self, queries: List[str], top_k: int, **kwargs
	) -> Tuple[List[List[str]], List[List[float]]]:
		queries = self.truncated_inputs(queries)
		query_embeddings: List[
			List[float]
		] = await self.embedding.aget_text_embedding_batch(queries)

		self.collection.load(timeout=self.timeout)

		# Perform similarity search
		results = self.collection.search(
			data=query_embeddings,
			limit=top_k,
			anns_field="vector",
			param={"metric_type": self.similarity_metric.upper()},
			timeout=self.timeout,
			**kwargs,
		)

		# Extract IDs and distances
		ids = [[str(hit.id) for hit in result] for result in results]
		distances = [[hit.distance for hit in result] for result in results]

		if self.similarity_metric in ["l2"]:
			distances = apply_recursive(lambda x: -x, distances)

		return ids, distances

	async def fetch(self, ids: List[str]) -> List[List[float]]:
		try:
			self.collection.load(timeout=self.timeout)
		except MilvusException as e:
			logger.warning(f"Failed to load collection: {e}")
			return [[]] * len(ids)
		# Fetch vectors by IDs
		results = self.collection.query(
			expr=f"id in {ids}", output_fields=["id", "vector"], timeout=self.timeout
		)
		id_vector_dict = {str(result["id"]): result["vector"] for result in results}
		result = [id_vector_dict[_id] for _id in ids]
		return result

	async def is_exist(self, ids: List[str]) -> List[bool]:
		try:
			self.collection.load(timeout=self.timeout)
		except MilvusException:
			return [False] * len(ids)
		# Check the existence of IDs
		results = self.collection.query(
			expr=f"id in {ids}", output_fields=["id"], timeout=self.timeout
		)
		# Determine existence
		existing_ids = {str(result["id"]) for result in results}
		return [str(_id) in existing_ids for _id in ids]

	async def delete(self, ids: List[str]):
		# Delete entries by IDs
		self.collection.delete(expr=f"id in {ids}", timeout=self.timeout)

	def delete_collection(self):
		# Delete the collection
		self.collection.release(timeout=self.timeout)
		self.collection.drop_index(timeout=self.timeout)
		self.collection.drop(timeout=self.timeout)



from typing import Dict


def add_gen_gt(row: Dict, new_gen_gt: str) -> Dict:
	if "generation_gt" in list(row.keys()):
		if isinstance(row["generation_gt"], list):
			row["generation_gt"].append(new_gen_gt)
		elif isinstance(row["generation_gt"], str):
			row["generation_gt"] = [row["generation_gt"], new_gen_gt]
		else:
			raise ValueError(
				"generation_gt should be either a string or a list of strings."
			)
		return row
	row["generation_gt"] = [new_gen_gt]
	return row



from typing import Optional

import click
import streamlit as st

from autorag.deploy import Runner


def get_runner(
	yaml_path: Optional[str], project_dir: Optional[str], trial_path: Optional[str]
):
	if not yaml_path and not trial_path:
		raise ValueError("yaml_path or trial_path must be given.")
	elif yaml_path and trial_path:
		raise ValueError("yaml_path and trial_path cannot be given at the same time.")
	elif yaml_path:
		return Runner.from_yaml(yaml_path, project_dir=project_dir)
	elif trial_path:
		return Runner.from_trial_folder(trial_path)


def set_initial_state():
	if "messages" not in st.session_state:
		st.session_state["messages"] = [
			{
				"role": "assistant",
				"content": "Welcome !",
			}
		]


def set_page_config():
	st.set_page_config(
		page_title="AutoRAG",
		page_icon="🤖",
		layout="wide",
		initial_sidebar_state="expanded",
		menu_items={
			"Get help": "https://github.com/Marker-Inc-Korea/AutoRAG/discussions",
			"Report a bug": "https://github.com/Marker-Inc-Korea/AutoRAG/issues",
		},
	)


def set_page_header():
	st.header("📚 AutoRAG", anchor=False)
	st.caption("Input a question and get an answer from the given documents. ")


def chat_box(runner: Runner):
	if query := st.chat_input("How can I help?"):
		# Add the user input to messages state
		st.session_state["messages"].append({"role": "user", "content": query})
		with st.chat_message("user"):
			st.markdown(query)

		# Generate llama-index stream with user input
		with st.chat_message("assistant"):
			with st.spinner("Processing..."):
				response = st.write(runner.run(query))

		# Add the final response to messages state
		st.session_state["messages"].append({"role": "assistant", "content": response})


@click.command()
@click.option("--yaml_path", type=str, help="Path to the YAML file.")
@click.option("--project_dir", type=str, help="Path to the project directory.")
@click.option("--trial_path", type=str, help="Path to the trial directory.")
def run_web_server(
	yaml_path: Optional[str], project_dir: Optional[str], trial_path: Optional[str]
):
	runner = get_runner(yaml_path, project_dir, trial_path)
	set_initial_state()
	set_page_config()
	set_page_header()
	chat_box(runner)


if __name__ == "__main__":
	run_web_server()



import os
from itertools import chain
import uuid
from typing import Tuple, List, Dict, Any, Optional

from langchain_text_splitters import TextSplitter

from autorag.data.chunk.base import chunker_node, add_file_name
from autorag.data.utils.util import add_essential_metadata, get_start_end_idx


@chunker_node
def langchain_chunk(
	texts: List[str],
	chunker: TextSplitter,
	file_name_language: Optional[str] = None,
	metadata_list: Optional[List[Dict[str, str]]] = None,
) -> Tuple[
	List[str], List[str], List[str], List[Tuple[int, int]], List[Dict[str, Any]]
]:
	"""
	Chunk texts from the parsed result to use langchain chunk method

	:param texts: The list of texts to chunk from the parsed result
	:param chunker: A langchain TextSplitter(Chunker) instance.
	:param file_name_language: The language to use 'add_file_name' feature.
	    You need to set one of 'English' and 'Korean'
	    The 'add_file_name' feature is to add a file_name to chunked_contents.
	    This is used to prevent hallucination by retrieving contents from the wrong document.
	    Default form of 'English' is "file_name: {file_name}\n contents: {content}"
	:param metadata_list: The list of dict of metadata from the parsed result
	:return: tuple of lists containing the chunked doc_id, contents, path, start_idx, end_idx and metadata
	"""
	results = [
		langchain_chunk_pure(text, chunker, file_name_language, meta)
		for text, meta in zip(texts, metadata_list)
	]

	doc_id, contents, path, start_end_idx, metadata = (
		list(chain.from_iterable(item)) for item in zip(*results)
	)

	return doc_id, contents, path, start_end_idx, metadata


def langchain_chunk_pure(
	text: str,
	chunker: TextSplitter,
	file_name_language: Optional[str] = None,
	_metadata: Optional[Dict[str, str]] = None,
):
	# chunk
	chunk_results = chunker.create_documents([text], metadatas=[_metadata])

	# make doc_id
	doc_id = list(str(uuid.uuid4()) for _ in range(len(chunk_results)))

	# make path
	path_lst = list(map(lambda x: x.metadata.get("path", ""), chunk_results))

	# make contents and start_end_idx
	if file_name_language:
		chunked_file_names = list(map(lambda x: os.path.basename(x), path_lst))
		chunked_texts = list(map(lambda x: x.page_content, chunk_results))
		start_end_idx = list(map(lambda x: get_start_end_idx(text, x), chunked_texts))
		contents = add_file_name(file_name_language, chunked_file_names, chunked_texts)
	else:
		contents = list(map(lambda node: node.page_content, chunk_results))
		start_end_idx = list(map(lambda x: get_start_end_idx(text, x), contents))

	# make metadata
	metadata = list(
		map(lambda node: add_essential_metadata(node.metadata), chunk_results)
	)

	return doc_id, contents, path_lst, start_end_idx, metadata



from typing import List

import pandas as pd

from autorag.nodes.queryexpansion.base import BaseQueryExpansion
from autorag.utils import result_to_dataframe

hyde_prompt = "Please write a passage to answer the question"


class HyDE(BaseQueryExpansion):
	@result_to_dataframe(["queries"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries = self.cast_to_run(previous_result, *args, **kwargs)

		# pop prompt from kwargs
		prompt = kwargs.pop("prompt", hyde_prompt)
		kwargs.pop("generator_module_type", None)

		expanded_queries = self._pure(queries, prompt, **kwargs)
		return self._check_expanded_query(queries, expanded_queries)

	def _pure(self, queries: List[str], prompt: str = hyde_prompt, **generator_params):
		"""
		HyDE, which inspired by "Precise Zero-shot Dense Retrieval without Relevance Labels" (https://arxiv.org/pdf/2212.10496.pdf)
		LLM model creates a hypothetical passage.
		And then, retrieve passages using hypothetical passage as a query.
		:param queries: List[str], queries to retrieve.
		:param prompt: Prompt to use when generating hypothetical passage
		:return: List[List[str]], List of hyde results.
		"""
		full_prompts = list(
			map(
				lambda x: (prompt if not bool(prompt) else hyde_prompt)
				+ f"\nQuestion: {x}\nPassage:",
				queries,
			)
		)
		input_df = pd.DataFrame({"prompts": full_prompts})
		result_df = self.generator.pure(previous_result=input_df, **generator_params)
		answers = result_df["generated_texts"].tolist()
		results = list(map(lambda x: [x], answers))
		return results



import gc
from copy import deepcopy
from typing import List, Tuple, Union

import pandas as pd

from autorag.nodes.generator.base import BaseGenerator
from autorag.utils import result_to_dataframe
from autorag.utils.util import pop_params, to_list, is_chat_prompt


class Vllm(BaseGenerator):
	def __init__(self, project_dir: str, llm: str, **kwargs):
		super().__init__(project_dir, llm, **kwargs)
		try:
			from vllm import SamplingParams, LLM
		except ImportError:
			raise ImportError(
				"Please install vllm library. You can install it by running `pip install vllm`."
			)

		model_from_kwargs = kwargs.pop("model", None)
		model = llm if model_from_kwargs is None else model_from_kwargs

		input_kwargs = deepcopy(kwargs)
		sampling_params_init_params = pop_params(
			SamplingParams.from_optional, input_kwargs
		)
		self.vllm_model = LLM(model, **input_kwargs)

		# delete not sampling param keys in the kwargs
		kwargs_keys = list(kwargs.keys())
		for key in kwargs_keys:
			if key not in sampling_params_init_params:
				kwargs.pop(key)

	def __del__(self):
		try:
			import torch
			import contextlib

			if torch.cuda.is_available():
				from vllm.distributed.parallel_state import (
					destroy_model_parallel,
					destroy_distributed_environment,
				)

				destroy_model_parallel()
				destroy_distributed_environment()
				if hasattr(self.vllm_model.llm_engine, 'model_executor'):
					del self.vllm_model.llm_engine.model_executor
				del self.vllm_model
				with contextlib.suppress(AssertionError):
					torch.distributed.destroy_process_group()
				gc.collect()
				torch.cuda.empty_cache()
				torch.cuda.synchronize()
		except ImportError:
			del self.vllm_model

		super().__del__()

	@result_to_dataframe(["generated_texts", "generated_tokens", "generated_log_probs"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		prompts = self.cast_to_run(previous_result)
		thinking = kwargs.pop("thinking", False)
		return self._pure(prompts, thinking=thinking, **kwargs)

	def _pure(
		self,
		prompts: Union[List[str], List[List[dict]]],
		thinking: bool = False,
		**kwargs,
	) -> Tuple[List[str], List[List[int]], List[List[float]]]:
		"""
		Vllm module.
		It gets the VLLM instance and returns generated texts by the input prompt.
		You can set logprobs to get the log probs of the generated text.
		Default logprobs is 1.

		:param prompts: A list of prompts or a list of chat prompts.
		:param thinking: A boolean that indicates whether to think when generating text.
			Default is False.
			Effective when set True and using chat prompts.
			You can learn how to use chat prompt at `chat_fstring` module documentation.
		:param kwargs: The extra parameters for generating the text.
		:return: A tuple of three elements.
		    The first element is a list of generated text.
		    The second element is a list of generated text's token ids.
		    The third element is a list of generated text's log probs.
		"""
		try:
			from vllm.outputs import RequestOutput
			from vllm.sequence import SampleLogprobs
			from vllm import SamplingParams, LLM
		except ImportError:
			raise ImportError(
				"Please install vllm library. You can install it by running `pip install vllm`."
			)

		if "logprobs" not in kwargs:
			kwargs["logprobs"] = 1

		sampling_params = pop_params(SamplingParams.from_optional, kwargs)
		generate_params = SamplingParams(**sampling_params)
		if is_chat_prompt(prompts):
			chat_template_kwargs = kwargs.pop("chat_template_kwargs", {})
			chat_template_kwargs["enable_thinking"] = thinking
			chat_kwargs = pop_params(LLM.chat, kwargs)
			results: List[RequestOutput] = self.vllm_model.chat(
				prompts,
				generate_params,
				chat_template_kwargs=chat_template_kwargs,
				**chat_kwargs,
			)
		else:
			generate_kwargs = pop_params(LLM.generate, kwargs)
			results: List[RequestOutput] = self.vllm_model.generate(
				prompts, generate_params, **generate_kwargs
			)
		generated_texts = list(map(lambda x: x.outputs[0].text, results))
		generated_token_ids = list(map(lambda x: x.outputs[0].token_ids, results))
		log_probs: List[SampleLogprobs] = list(
			map(lambda x: x.outputs[0].logprobs, results)
		)
		generated_log_probs = list(
			map(
				lambda x: list(map(lambda y: y[0][y[1]].logprob, zip(x[0], x[1]))),
				zip(log_probs, generated_token_ids),
			)
		)
		return (
			to_list(generated_texts),
			to_list(generated_token_ids),
			to_list(generated_log_probs),
		)

	async def astream(self, prompt: str, **kwargs):
		raise NotImplementedError

	def stream(self, prompt: str, **kwargs):
		raise NotImplementedError



from typing import List

import pandas as pd

from autorag.nodes.passageaugmenter.base import BasePassageAugmenter
from autorag.utils import result_to_dataframe


class PassPassageAugmenter(BasePassageAugmenter):
	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		"""
		Run the passage augmenter node - PassPassageAugmenter module.

		:param previous_result: The previous result Dataframe.
		:param top_k: You must input the top_k value to get the top k results.
		:param kwargs: Not affected.
		:return: DataFrame with retrieved_contents, retrieved_ids, and retrieve_scores columns
		"""
		top_k = kwargs.pop("top_k")

		ids = self.cast_to_run(previous_result)
		contents = previous_result["retrieved_contents"].tolist()
		scores = previous_result["retrieve_scores"].tolist()

		augmented_ids, augmented_contents, augmented_scores = self._pure(
			ids, contents, scores
		)
		return self.sort_by_scores(
			augmented_contents, augmented_ids, augmented_scores, top_k
		)

	def _pure(
		self,
		ids_list: List[List[str]],
		contents_list: List[List[str]],
		scores_list: List[List[float]],
	):
		"""
		Do not perform augmentation.
		Return given passages, scores, and ids as is.
		"""
		return ids_list, contents_list, scores_list



from typing import List, Tuple

import pandas as pd
from llama_index.core.base.llms.base import BaseLLM
from transformers import AutoTokenizer

from autorag import generator_models
from autorag.nodes.generator.base import BaseGenerator
from autorag.utils.util import (
	get_event_loop,
	process_batch,
	result_to_dataframe,
	pop_params,
)


class LlamaIndexLLM(BaseGenerator):
	def __init__(self, project_dir: str, llm: str, batch: int = 16, *args, **kwargs):
		"""
		Initialize the Llama Index LLM module.

		:param project_dir: The project directory.
		:param llm: A llama index LLM instance.
		:param batch: The batch size for llm.
			Set low if you face some errors.
			Default is 16.
		:param kwargs: The extra parameters for initializing the llm instance.
		"""
		super().__init__(project_dir=project_dir, llm=llm)
		if self.llm not in generator_models.keys():
			raise ValueError(
				f"{self.llm} is not a valid llm name. Please check the llm name."
				"You can check valid llm names from autorag.generator_models."
			)
		self.batch = batch
		llm_class = generator_models[self.llm]

		if llm_class.class_name() in [
			"HuggingFace_LLM",
			"HuggingFaceInferenceAPI",
			"TextGenerationInference",
		]:
			model_name = kwargs.pop("model", None)
			if model_name is not None:
				kwargs["model_name"] = model_name
			else:
				if "model_name" not in kwargs.keys():
					raise ValueError(
						"`model` or `model_name` parameter must be provided for using huggingfacellm."
					)
			kwargs["tokenizer_name"] = kwargs["model_name"]
		self.llm_instance: BaseLLM = llm_class(**pop_params(llm_class.__init__, kwargs))

	def __del__(self):
		super().__del__()
		del self.llm_instance

	@result_to_dataframe(["generated_texts", "generated_tokens", "generated_log_probs"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		prompts = self.cast_to_run(previous_result=previous_result)
		return self._pure(prompts)

	def _pure(
		self,
		prompts: List[str],
	) -> Tuple[List[str], List[List[int]], List[List[float]]]:
		"""
		Llama Index LLM module.
		It gets the LLM instance from llama index, and returns generated text by the input prompt.
		It does not generate the right log probs, but it returns the pseudo log probs,
		which are not meant to be used for other modules.

		:param prompts: A list of prompts.
		:return: A tuple of three elements.
			The first element is a list of a generated text.
			The second element is a list of generated text's token ids, used tokenizer is GPT2Tokenizer.
			The third element is a list of generated text's pseudo log probs.
		"""
		tasks = [self.llm_instance.acomplete(prompt) for prompt in prompts]
		loop = get_event_loop()
		results = loop.run_until_complete(process_batch(tasks, batch_size=self.batch))

		generated_texts = list(map(lambda x: x.text, results))
		tokenizer = AutoTokenizer.from_pretrained("gpt2", use_fast=False)
		tokenized_ids = tokenizer(generated_texts).data["input_ids"]
		pseudo_log_probs = list(map(lambda x: [0.5] * len(x), tokenized_ids))
		return generated_texts, tokenized_ids, pseudo_log_probs

	async def astream(self, prompt: str, **kwargs):
		async for completion_response in await self.llm_instance.astream_complete(
			prompt
		):
			yield completion_response.text

	def stream(self, prompt: str, **kwargs):
		for completion_response in self.llm_instance.stream_complete(prompt):
			yield completion_response.text



import itertools
from typing import Dict, List

from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.base.llms.types import ChatResponse, ChatMessage, MessageRole

from autorag.data.qa.evolve.prompt import QUERY_EVOLVE_PROMPT


async def llama_index_generate_base(
	row: Dict,
	llm: BaseLLM,
	messages: List[ChatMessage],
) -> Dict:
	original_query = row["query"]
	context = list(itertools.chain.from_iterable(row["retrieval_gt_contents"]))
	context_str = "Text:\n" + "\n".join(
		[f"{i + 1}. {c}" for i, c in enumerate(context)]
	)
	user_prompt = f"Question: {original_query}\nContext: {context_str}\nOutput: "
	messages.append(ChatMessage(role=MessageRole.USER, content=user_prompt))

	chat_response: ChatResponse = await llm.achat(messages=messages)
	row["query"] = chat_response.message.content
	return row


async def conditional_evolve_ragas(
	row: Dict,
	llm: BaseLLM,
	lang: str = "en",
) -> Dict:
	return await llama_index_generate_base(
		row,
		llm,
		QUERY_EVOLVE_PROMPT["conditional_evolve_ragas"][lang],
	)


async def reasoning_evolve_ragas(
	row: Dict,
	llm: BaseLLM,
	lang: str = "en",
) -> Dict:
	return await llama_index_generate_base(
		row,
		llm,
		QUERY_EVOLVE_PROMPT["reasoning_evolve_ragas"][lang],
	)


async def compress_ragas(
	row: Dict,
	llm: BaseLLM,
	lang: str = "en",
) -> Dict:
	original_query = row["query"]
	user_prompt = f"Question: {original_query}\nOutput: "
	messages = QUERY_EVOLVE_PROMPT["compress_ragas"][lang]
	messages.append(ChatMessage(role=MessageRole.USER, content=user_prompt))

	chat_response: ChatResponse = await llm.achat(messages=messages)
	row["query"] = chat_response.message.content
	return row



import abc
import logging
import os

import pandas as pd

from autorag.schema import BaseModule
from autorag.utils import (
	validate_qa_dataset,
	sort_by_scores,
	validate_corpus_dataset,
	cast_corpus_dataset,
)
from autorag.utils.cast import cast_retrieved_ids
from autorag.utils.util import select_top_k

logger = logging.getLogger("AutoRAG")


class BasePassageAugmenter(BaseModule, metaclass=abc.ABCMeta):
	def __init__(self, project_dir: str, *args, **kwargs):
		logger.info(
			f"Initialize passage augmenter node - {self.__class__.__name__} module..."
		)
		data_dir = os.path.join(project_dir, "data")
		corpus_df = pd.read_parquet(
			os.path.join(data_dir, "corpus.parquet"), engine="pyarrow"
		)
		validate_corpus_dataset(corpus_df)
		corpus_df = cast_corpus_dataset(corpus_df)
		self.corpus_df = corpus_df

	def __del__(self):
		logger.info(
			f"Initialize passage augmenter node - {self.__class__.__name__} module..."
		)

	def cast_to_run(self, previous_result: pd.DataFrame, *args, **kwargs):
		logger.info(
			f"Running passage augmenter node - {self.__class__.__name__} module..."
		)
		validate_qa_dataset(previous_result)

		# find ids columns
		return cast_retrieved_ids(previous_result)

	@staticmethod
	def sort_by_scores(
		augmented_contents,
		augmented_ids,
		augmented_scores,
		top_k: int,
		reverse: bool = True,
	):
		# sort by scores
		df = pd.DataFrame(
			{
				"contents": augmented_contents,
				"ids": augmented_ids,
				"scores": augmented_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			lambda row: sort_by_scores(row, reverse=reverse),
			axis=1,
			result_type="expand",
		)

		# select by top_k
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)



GEN_GT_SYSTEM_PROMPT = {
	"concise": {
		"en": """You are an AI assistant to answer the given question in the provide evidence text.
You can find the evidence from the given text about question, and you have to write a proper answer to the given question.
Your answer have to be concise and relevant to the question.
Do not make a verbose answer and make it super clear.
It doesn't have to be an full sentence. It can be the answer is a word or a paraphrase.""",
		"ko": """당신은 주어진 질문에 대해 제공된 Text 내에서 답을 찾는 AI 비서입니다.
질문에 대한 답을 Text에서 찾아 적절한 답변을 작성하세요.
답변은 간결하고 질문에 관련된 내용만 포함해야 합니다.
불필요하게 길게 답변하지 말고, 명확하게 작성하세요.
완전한 문장이 아니어도 되며, 답은 단어나 요약일 수 있습니다.""",
		"ja": """
あなたは与えられた質問に対して提供されたText内で答えを探すAI秘書です。
質問に対する答えをTextで探して適切な答えを作成しましょう。
回答は簡潔で、質問に関連する内容のみを含める必要があります。
不必要に長く答えず、明確に作成しましょう。
完全な文章でなくてもいいし、答えは単語や要約かもしれません。
""",
	},
	"basic": {
		"en": """You are an AI assistant to answer the given question in the provide evidence text.
You can find the evidence from the given text about question, and you have to write a proper answer to the given question.""",
		"ko": "당신은 주어진 질문에 대한 답을 제공된 Text 내에서 찾는 AI 비서입니다. 질문과 관련된 증거를 Text에서 찾아 적절한 답변을 작성하세요.",
		"ja": "あなたは与えられた質問に対する答えを提供されたText内で探すAI秘書です。 質問に関する証拠をTextで探して適切な回答を作成しましょう。",
	},
}



import asyncio
import os
import pickle
import re
from typing import List, Dict, Tuple, Callable, Union, Iterable, Optional

import numpy as np
import pandas as pd
from llama_index.core.indices.keyword_table.utils import simple_extract_keywords
from nltk import PorterStemmer
from rank_bm25 import BM25Okapi
from transformers import AutoTokenizer, PreTrainedTokenizerBase

from autorag.nodes.retrieval.base import (
	evenly_distribute_passages,
	BaseRetrieval,
	get_bm25_pkl_name,
)
from autorag.utils import validate_corpus_dataset, fetch_contents
from autorag.utils.util import (
	get_event_loop,
	normalize_string,
	result_to_dataframe,
	pop_params,
)


def tokenize_ko_kiwi(texts: List[str]) -> List[List[str]]:
	try:
		from kiwipiepy import Kiwi, Token
	except ImportError:
		raise ImportError(
			"You need to install kiwipiepy to use 'ko_kiwi' tokenizer. "
			"Please install kiwipiepy by running 'pip install kiwipiepy'. "
			"Or install Korean version of AutoRAG by running 'pip install AutoRAG[ko]'."
		)
	texts = list(map(lambda x: x.strip().lower(), texts))
	kiwi = Kiwi()
	tokenized_list: Iterable[List[Token]] = kiwi.tokenize(texts)

	def extract_form_safe(x):
		try:
			return x.form
		except UnicodeDecodeError:
			return " "

	return [
		list(map(lambda x: extract_form_safe(x), token_list))
		for token_list in tokenized_list
	]


def tokenize_ko_kkma(texts: List[str]) -> List[List[str]]:
	try:
		from konlpy.tag import Kkma
	except ImportError:
		raise ImportError(
			"You need to install konlpy to use 'ko_kkma' tokenizer. "
			"Please install konlpy by running 'pip install konlpy'. "
			"Or install Korean version of AutoRAG by running 'pip install AutoRAG[ko]'."
		)
	tokenizer = Kkma()
	tokenized_list: List[List[str]] = list(map(lambda x: tokenizer.morphs(x), texts))
	return tokenized_list


def tokenize_ko_okt(texts: List[str]) -> List[List[str]]:
	try:
		from konlpy.tag import Okt
	except ImportError:
		raise ImportError(
			"You need to install konlpy to use 'ko_kkma' tokenizer. "
			"Please install konlpy by running 'pip install konlpy'. "
			"Or install Korean version of AutoRAG by running 'pip install AutoRAG[ko]'."
		)
	tokenizer = Okt()
	tokenized_list: List[List[str]] = list(map(lambda x: tokenizer.morphs(x), texts))
	return tokenized_list


def tokenize_porter_stemmer(texts: List[str]) -> List[List[str]]:
	def tokenize_remove_stopword(text: str, stemmer) -> List[str]:
		text = text.lower()
		words = list(simple_extract_keywords(text))
		return [stemmer.stem(word) for word in words]

	stemmer = PorterStemmer()
	tokenized_list: List[List[str]] = list(
		map(lambda x: tokenize_remove_stopword(x, stemmer), texts)
	)
	return tokenized_list


def tokenize_space(texts: List[str]) -> List[List[str]]:
	def tokenize_space_text(text: str) -> List[str]:
		text = normalize_string(text)
		return re.split(r"\s+", text.strip())

	return list(map(tokenize_space_text, texts))


def load_bm25_corpus(bm25_path: str) -> Dict:
	if bm25_path is None:
		return {}
	with open(bm25_path, "rb") as f:
		bm25_corpus = pickle.load(f)
	return bm25_corpus


def tokenize_ja_sudachipy(texts: List[str]) -> List[List[str]]:
	try:
		from sudachipy import dictionary, tokenizer
	except ImportError:
		raise ImportError(
			"You need to install SudachiPy to use 'sudachipy' tokenizer. "
			"Please install SudachiPy by running 'pip install sudachipy'."
		)

	# Initialize SudachiPy with the default tokenizer
	tokenizer_obj = dictionary.Dictionary(dict="core").create()

	# Choose the tokenizer mode: NORMAL, SEARCH, A
	mode = tokenizer.Tokenizer.SplitMode.A

	# Tokenize the input texts
	tokenized_list = []
	for text in texts:
		tokens = tokenizer_obj.tokenize(text, mode)
		tokenized_list.append([token.surface() for token in tokens])

	return tokenized_list


BM25_TOKENIZER = {
	"porter_stemmer": tokenize_porter_stemmer,
	"ko_kiwi": tokenize_ko_kiwi,
	"space": tokenize_space,
	"ko_kkma": tokenize_ko_kkma,
	"ko_okt": tokenize_ko_okt,
	"sudachipy": tokenize_ja_sudachipy,
}


class BM25(BaseRetrieval):
	def __init__(self, project_dir: str, *args, **kwargs):
		"""
		Initialize BM25 module.
		(Retrieval)

		:param project_dir: The project directory path.
		:param bm25_tokenizer: The tokenizer name that is used to the BM25.
		    It supports 'porter_stemmer', 'ko_kiwi', and huggingface `AutoTokenizer`.
		    You can pass huggingface tokenizer name.
		    Default is porter_stemmer.
		:param kwargs: The optional arguments.
		"""

		super().__init__(project_dir)
		# check if bm25_path and file exist
		bm25_tokenizer = kwargs.get("bm25_tokenizer", None)
		if bm25_tokenizer is None:
			bm25_tokenizer = "porter_stemmer"
		bm25_path = os.path.join(self.resources_dir, get_bm25_pkl_name(bm25_tokenizer))

		assert (
			bm25_path is not None
		), "bm25_path must be specified for using bm25 retrieval."
		assert os.path.exists(
			bm25_path
		), f"bm25_path {bm25_path} does not exist. Please ingest first."

		self.bm25_corpus = load_bm25_corpus(bm25_path)
		assert (
			"tokens" and "passage_id" in list(self.bm25_corpus.keys())
		), "bm25_corpus must contain tokens and passage_id. Please check you ingested bm25 corpus correctly."
		self.tokenizer = select_bm25_tokenizer(bm25_tokenizer)
		assert self.bm25_corpus["tokenizer_name"] == bm25_tokenizer, (
			f"The bm25 corpus tokenizer is {self.bm25_corpus['tokenizer_name']}, but your input is {bm25_tokenizer}. "
			f"You need to ingest again. Delete bm25 pkl file and re-ingest it."
		)
		self.bm25_instance = BM25Okapi(self.bm25_corpus["tokens"])

	@result_to_dataframe(
		[
			"retrieved_contents_lexical",
			"retrieved_ids_lexical",
			"retrieve_scores_lexical",
		]
	)
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries = self.cast_to_run(previous_result)
		pure_params = pop_params(self._pure, kwargs)
		ids, scores = self._pure(queries, *args, **pure_params)
		contents = fetch_contents(self.corpus_df, ids)
		return contents, ids, scores

	def _pure(
		self,
		queries: List[List[str]],
		top_k: int,
		ids: Optional[List[List[str]]] = None,
	) -> Tuple[List[List[str]], List[List[float]]]:
		"""
		BM25 retrieval function.
		You have to load a pickle file that is already ingested.

		:param queries: 2-d list of query strings.
		    Each element of the list is a query strings of each row.
		:param top_k: The number of passages to be retrieved.
		:param ids: The optional list of ids that you want to retrieve.
		    You don't need to specify this in the general use cases.
		    Default is None.
		:return: The 2-d list contains a list of passage ids that retrieved from bm25 and 2-d list of its scores.
		    It will be a length of queries. And each element has a length of top_k.
		"""
		if ids is not None:
			score_result = list(
				map(
					lambda query_list, id_list: get_bm25_scores(
						query_list,
						id_list,
						self.tokenizer,
						self.bm25_instance,
						self.bm25_corpus,
					),
					queries,
					ids,
				)
			)
			return ids, score_result

		# run async bm25_pure function
		tasks = [
			bm25_pure(
				input_queries,
				top_k,
				self.tokenizer,
				self.bm25_instance,
				self.bm25_corpus,
			)
			for input_queries in queries
		]
		loop = get_event_loop()
		results = loop.run_until_complete(asyncio.gather(*tasks))
		id_result = list(map(lambda x: x[0], results))
		score_result = list(map(lambda x: x[1], results))
		return id_result, score_result


async def bm25_pure(
	queries: List[str], top_k: int, tokenizer, bm25_api: BM25Okapi, bm25_corpus: Dict
) -> Tuple[List[str], List[float]]:
	"""
	Async BM25 retrieval function.
	Its usage is for async retrieval of bm25 row by row.

	:param queries: A list of query strings.
	:param top_k: The number of passages to be retrieved.
	:param tokenizer: A tokenizer that will be used to tokenize queries.
	:param bm25_api: A bm25 api instance that will be used to retrieve passages.
	:param bm25_corpus: A dictionary containing the bm25 corpus, which is doc_id from corpus and tokenized corpus.
	    Its data structure looks like this:

	    .. Code:: python

	        {
	            "tokens": [], # 2d list of tokens
	            "passage_id": [], # 2d list of passage_id. Type must be str.
	        }
	:return: The tuple contains a list of passage ids that retrieved from bm25 and its scores.
	"""
	# I don't make queries operation to async, because queries length might be small, so it will occur overhead.
	tokenized_queries = tokenize(queries, tokenizer)
	id_result = []
	score_result = []
	for query in tokenized_queries:
		scores = bm25_api.get_scores(query)
		sorted_scores = sorted(scores, reverse=True)
		top_n_index = np.argsort(scores)[::-1][:top_k]
		ids = [bm25_corpus["passage_id"][i] for i in top_n_index]
		id_result.append(ids)
		score_result.append(sorted_scores[:top_k])

	# make a total result to top_k
	id_result, score_result = evenly_distribute_passages(id_result, score_result, top_k)
	# sort id_result and score_result by score
	result = [
		(_id, score)
		for score, _id in sorted(
			zip(score_result, id_result), key=lambda pair: pair[0], reverse=True
		)
	]
	id_result, score_result = zip(*result)
	return list(id_result), list(score_result)


def get_bm25_scores(
	queries: List[str],
	ids: List[str],
	tokenizer,
	bm25_api: BM25Okapi,
	bm25_corpus: Dict,
) -> List[float]:
	if len(ids) == 0 or not bool(ids):
		return []
	tokenized_queries = tokenize(queries, tokenizer)
	result_dict = {id_: [] for id_ in ids}
	for query in tokenized_queries:
		scores = bm25_api.get_scores(query)
		for i, id_ in enumerate(ids):
			result_dict[id_].append(scores[bm25_corpus["passage_id"].index(id_)])
	result_df = pd.DataFrame(result_dict)
	return result_df.max(axis=0).tolist()


def tokenize(queries: List[str], tokenizer) -> List[List[int]]:
	if isinstance(tokenizer, PreTrainedTokenizerBase):
		tokenized_queries = tokenizer(queries).input_ids
	else:
		tokenized_queries = tokenizer(queries)
	return tokenized_queries


def bm25_ingest(
	corpus_path: str, corpus_data: pd.DataFrame, bm25_tokenizer: str = "porter_stemmer"
):
	if not corpus_path.endswith(".pkl"):
		raise ValueError(f"Corpus path {corpus_path} is not a pickle file.")
	validate_corpus_dataset(corpus_data)
	ids = corpus_data["doc_id"].tolist()

	# Initialize bm25_corpus
	bm25_corpus = pd.DataFrame()

	# Load the BM25 corpus if it exists and get the passage ids
	if os.path.exists(corpus_path) and os.path.getsize(corpus_path) > 0:
		with open(corpus_path, "rb") as r:
			corpus = pickle.load(r)
			bm25_corpus = pd.DataFrame.from_dict(corpus)
		duplicated_passage_rows = bm25_corpus[bm25_corpus["passage_id"].isin(ids)]
		new_passage = corpus_data[
			~corpus_data["doc_id"].isin(duplicated_passage_rows["passage_id"])
		]
	else:
		new_passage = corpus_data

	if not new_passage.empty:
		tokenizer = select_bm25_tokenizer(bm25_tokenizer)
		if isinstance(tokenizer, PreTrainedTokenizerBase):
			tokenized_corpus = tokenizer(new_passage["contents"].tolist()).input_ids
		else:
			tokenized_corpus = tokenizer(new_passage["contents"].tolist())
		new_bm25_corpus = pd.DataFrame(
			{
				"tokens": tokenized_corpus,
				"passage_id": new_passage["doc_id"].tolist(),
			}
		)

		if not bm25_corpus.empty:
			bm25_corpus_updated = pd.concat(
				[bm25_corpus, new_bm25_corpus], ignore_index=True
			)
			bm25_dict = bm25_corpus_updated.to_dict("list")
		else:
			bm25_dict = new_bm25_corpus.to_dict("list")

		# add tokenizer name to bm25_dict
		bm25_dict["tokenizer_name"] = bm25_tokenizer

		with open(corpus_path, "wb") as w:
			pickle.dump(bm25_dict, w)


def select_bm25_tokenizer(
	bm25_tokenizer: str,
) -> Callable[[str], List[Union[int, str]]]:
	if bm25_tokenizer in list(BM25_TOKENIZER.keys()):
		return BM25_TOKENIZER[bm25_tokenizer]

	return AutoTokenizer.from_pretrained(bm25_tokenizer, use_fast=False)



import itertools
from typing import Dict


from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.base.llms.types import MessageRole, ChatMessage

from autorag.data.qa.generation_gt.base import add_gen_gt
from autorag.data.qa.generation_gt.prompt import GEN_GT_SYSTEM_PROMPT


async def make_gen_gt_llama_index(row: Dict, llm: BaseLLM, system_prompt: str) -> Dict:
	retrieval_gt_contents = list(
		itertools.chain.from_iterable(row["retrieval_gt_contents"])
	)
	query = row["query"]
	passage_str = "\n".join(retrieval_gt_contents)
	user_prompt = f"Text:\n<|text_start|>\n{passage_str}\n<|text_end|>\n\nQuestion:\n{query}\n\nAnswer:"

	response = await llm.achat(
		messages=[
			ChatMessage(role=MessageRole.SYSTEM, content=system_prompt),
			ChatMessage(role=MessageRole.USER, content=user_prompt),
		],
		temperature=0.0,
	)
	return add_gen_gt(row, response.message.content)


async def make_concise_gen_gt(row: Dict, llm: BaseLLM, lang: str = "en") -> Dict:
	return await make_gen_gt_llama_index(
		row, llm, GEN_GT_SYSTEM_PROMPT["concise"][lang]
	)


async def make_basic_gen_gt(row: Dict, llm: BaseLLM, lang: str = "en") -> Dict:
	return await make_gen_gt_llama_index(row, llm, GEN_GT_SYSTEM_PROMPT["basic"][lang])


async def make_custom_gen_gt(row: Dict, llm: BaseLLM, system_prompt: str) -> Dict:
	return await make_gen_gt_llama_index(row, llm, system_prompt)



import logging
from abc import ABCMeta
from pathlib import Path
from typing import Union

import pandas as pd

from autorag.schema.base import BaseModule
from autorag.utils.cast import cast_retrieved_contents

logger = logging.getLogger("AutoRAG")


class BasePromptMaker(BaseModule, metaclass=ABCMeta):
	def __init__(self, project_dir: Union[str, Path], *args, **kwargs):
		logger.info(
			f"Initialize prompt maker node - {self.__class__.__name__} module..."
		)

	def __del__(self):
		logger.info(f"Prompt maker node - {self.__class__.__name__} module is deleted.")

	def cast_to_run(self, previous_result: pd.DataFrame, *args, **kwargs):
		logger.info(f"Running prompt maker node - {self.__class__.__name__} module...")
		# get query and retrieved contents from previous_result
		assert (
			"query" in previous_result.columns
		), "previous_result must have query column."

		query = previous_result["query"].tolist()
		prompt = kwargs.pop("prompt")
		return query, cast_retrieved_contents(previous_result), prompt



from typing import List

import pandas as pd

from autorag.nodes.queryexpansion.base import BaseQueryExpansion
from autorag.utils import result_to_dataframe

decompose_prompt = """Decompose a question in self-contained sub-questions. Use \"The question needs no decomposition\" when no decomposition is needed.

    Example 1:

    Question: Is Hamlet more common on IMDB than Comedy of Errors?
    Decompositions:
    1: How many listings of Hamlet are there on IMDB?
    2: How many listing of Comedy of Errors is there on IMDB?

    Example 2:

    Question: Are birds important to badminton?

    Decompositions:
    The question needs no decomposition

    Example 3:

    Question: Is it legal for a licensed child driving Mercedes-Benz to be employed in US?

    Decompositions:
    1: What is the minimum driving age in the US?
    2: What is the minimum age for someone to be employed in the US?

    Example 4:

    Question: Are all cucumbers the same texture?

    Decompositions:
    The question needs no decomposition

    Example 5:

    Question: Hydrogen's atomic number squared exceeds number of Spice Girls?

    Decompositions:
    1: What is the atomic number of hydrogen?
    2: How many Spice Girls are there?

    Example 6:

    Question: {question}

    Decompositions:
    """


class QueryDecompose(BaseQueryExpansion):
	@result_to_dataframe(["queries"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries = self.cast_to_run(previous_result, *args, **kwargs)

		# pop prompt from kwargs
		prompt = kwargs.pop("prompt", decompose_prompt)
		kwargs.pop("generator_module_type", None)

		expanded_queries = self._pure(queries, prompt, **kwargs)
		return self._check_expanded_query(queries, expanded_queries)

	def _pure(
		self, queries: List[str], prompt: str = decompose_prompt, *args, **kwargs
	) -> List[List[str]]:
		"""
		decompose query to little piece of questions.
		:param queries: List[str], queries to decompose.
		:param prompt: str, prompt to use for query decomposition.
			default prompt comes from Visconde's StrategyQA few-shot prompt.
		:return: List[List[str]], list of decomposed query. Return input query if query is not decomposable.
		"""
		full_prompts = []
		for query in queries:
			if bool(prompt):
				full_prompt = f"prompt: {prompt}\n\n question: {query}"
			else:
				full_prompt = decompose_prompt.format(question=query)
			full_prompts.append(full_prompt)
		input_df = pd.DataFrame({"prompts": full_prompts})
		result_df = self.generator.pure(previous_result=input_df, *args, **kwargs)
		answers = result_df["generated_texts"].tolist()
		results = list(
			map(lambda x: get_query_decompose(x[0], x[1]), zip(queries, answers))
		)
		return results


def get_query_decompose(query: str, answer: str) -> List[str]:
	"""
	decompose query to little piece of questions.
	:param query: str, query to decompose.
	:param answer: str, answer from query_decompose function.
	:return: List[str], list of a decomposed query. Return input query if query is not decomposable.
	"""
	if answer.lower() == "the question needs no decomposition":
		return [query]
	try:
		lines = [line.strip() for line in answer.splitlines() if line.strip()]
		if lines[0].startswith("Decompositions:"):
			lines.pop(0)
		questions = [line.split(":", 1)[1].strip() for line in lines if ":" in line]
		if not questions:
			return [query]
		return questions
	except:
		return [query]



from typing import List, Tuple

import numpy as np
import pandas as pd

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import (
	flatten_apply,
	sort_by_scores,
	select_top_k,
	pop_params,
	result_to_dataframe,
	empty_cuda_cache,
)


class ColbertReranker(BasePassageReranker):
	def __init__(
		self,
		project_dir: str,
		model_name: str = "colbert-ir/colbertv2.0",
		*args,
		**kwargs,
	):
		"""
		Initialize a colbert rerank model for reranking.

		:param project_dir: The project directory
		:param model_name: The model name for Colbert rerank.
			You can choose a colbert model for reranking.
			The default is "colbert-ir/colbertv2.0".
		:param kwargs: Extra parameter for the model.
		"""
		super().__init__(project_dir)
		try:
			import torch
			from transformers import AutoModel, AutoTokenizer
		except ImportError:
			raise ImportError(
				"Pytorch is not installed. Please install pytorch to use Colbert reranker."
			)
		self.device = "cuda" if torch.cuda.is_available() else "cpu"
		model_params = pop_params(AutoModel.from_pretrained, kwargs)
		self.model = AutoModel.from_pretrained(model_name, **model_params).to(
			self.device
		)
		self.tokenizer = AutoTokenizer.from_pretrained(model_name)

	def __del__(self):
		del self.model
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		batch = kwargs.pop("batch", 64)
		return self._pure(queries, contents, ids, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents with Colbert rerank models.
		You can get more information about a Colbert model at https://huggingface.co/colbert-ir/colbertv2.0.
		It uses BERT-based model, so recommend using CUDA gpu for faster reranking.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param batch: The number of queries to be processed in a batch
			Default is 64.

		:return: Tuple of lists containing the reranked contents, ids, and scores
		"""

		# get query and content embeddings
		query_embedding_list = get_colbert_embedding_batch(
			queries, self.model, self.tokenizer, batch
		)
		content_embedding_list = flatten_apply(
			get_colbert_embedding_batch,
			contents_list,
			model=self.model,
			tokenizer=self.tokenizer,
			batch_size=batch,
		)
		df = pd.DataFrame(
			{
				"ids": ids_list,
				"query_embedding": query_embedding_list,
				"contents": contents_list,
				"content_embedding": content_embedding_list,
			}
		)
		temp_df = df.explode("content_embedding")
		temp_df["score"] = temp_df.apply(
			lambda x: get_colbert_score(x["query_embedding"], x["content_embedding"]),
			axis=1,
		)
		df["scores"] = (
			temp_df.groupby(level=0, sort=False)["score"].apply(list).tolist()
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def get_colbert_embedding_batch(
	input_strings: List[str], model, tokenizer, batch_size: int
) -> List[np.array]:
	try:
		import torch
	except ImportError:
		raise ImportError(
			"Pytorch is not installed. Please install pytorch to use Colbert reranker."
		)
	encoding = tokenizer(
		input_strings,
		return_tensors="pt",
		padding=True,
		truncation=True,
		max_length=model.config.max_position_embeddings,
	)

	input_batches = slice_tokenizer_result(encoding, batch_size)
	result_embedding = []
	with torch.no_grad():
		for encoding_batch in input_batches:
			result_embedding.append(model(**encoding_batch).last_hidden_state)
	total_tensor = torch.cat(
		result_embedding, dim=0
	)  # shape [batch_size, token_length, embedding_dim]
	tensor_results = list(total_tensor.chunk(total_tensor.size()[0]))

	if torch.cuda.is_available():
		return list(map(lambda x: x.detach().cpu().numpy(), tensor_results))
	else:
		return list(map(lambda x: x.detach().numpy(), tensor_results))


def slice_tokenizer_result(tokenizer_output, batch_size):
	input_ids_batches = slice_tensor(tokenizer_output["input_ids"], batch_size)
	attention_mask_batches = slice_tensor(
		tokenizer_output["attention_mask"], batch_size
	)
	token_type_ids_batches = slice_tensor(
		tokenizer_output.get("token_type_ids", None), batch_size
	)
	return [
		{
			"input_ids": input_ids,
			"attention_mask": attention_mask,
			"token_type_ids": token_type_ids,
		}
		for input_ids, attention_mask, token_type_ids in zip(
			input_ids_batches, attention_mask_batches, token_type_ids_batches
		)
	]


def slice_tensor(input_tensor, batch_size):
	try:
		import torch
	except ImportError:
		raise ImportError(
			"Pytorch is not installed. Please install pytorch to use Colbert reranker."
		)
	# Calculate the number of full batches
	num_full_batches = input_tensor.size(0) // batch_size

	# Slice the tensor into batches
	tensor_list = [
		input_tensor[i * batch_size : (i + 1) * batch_size]
		for i in range(num_full_batches)
	]

	# Handle the last batch if it's smaller than batch_size
	remainder = input_tensor.size(0) % batch_size
	if remainder:
		tensor_list.append(input_tensor[-remainder:])

	device = "cuda" if torch.cuda.is_available() else "cpu"
	tensor_list = list(map(lambda x: x.to(device), tensor_list))

	return tensor_list


def get_colbert_score(query_embedding: np.array, content_embedding: np.array) -> float:
	if query_embedding.ndim == 3 and content_embedding.ndim == 3:
		query_embedding = query_embedding.reshape(-1, query_embedding.shape[-1])
		content_embedding = content_embedding.reshape(-1, content_embedding.shape[-1])

	sim_matrix = np.dot(query_embedding, content_embedding.T) / (
		np.linalg.norm(query_embedding, axis=1)[:, np.newaxis]
		* np.linalg.norm(content_embedding, axis=1)
	)
	max_sim_scores = np.max(sim_matrix, axis=1)
	return float(np.mean(max_sim_scores))



import logging
import os
import pathlib
from copy import deepcopy
from typing import List, Dict, Optional, Union

import pandas as pd

from autorag.evaluation import evaluate_retrieval
from autorag.schema.metricinput import MetricInput
from autorag.strategy import measure_speed, filter_by_threshold, select_best
from autorag.support import get_support_modules
from autorag.utils.cast import cast_retrieve_infos
from autorag.utils.util import make_combinations, explode

logger = logging.getLogger("AutoRAG")


def run_query_expansion_node(
	modules: List,
	module_params: List[Dict],
	previous_result: pd.DataFrame,
	node_line_dir: str,
	strategies: Dict,
) -> pd.DataFrame:
	"""
	Run evaluation and select the best module among query expansion node results.
	Initially, retrieval is run using expanded_queries, the result of the query_expansion module.
	The retrieval module is run as a combination of the retrieval_modules in strategies.
	If there are multiple retrieval_modules, run them all and choose the best result.
	If there are no retrieval_modules, run them with the default of bm25.
	In this way, the best result is selected for each module, and then the best result is selected.

	:param modules: Query expansion modules to run.
	:param module_params: Query expansion module parameters.
	:param previous_result: Previous result dataframe.
	    In this case, it would be qa data.
	:param node_line_dir: This node line's directory.
	:param strategies: Strategies for query expansion node.
	:return: The best result dataframe.
	"""
	if not os.path.exists(node_line_dir):
		os.makedirs(node_line_dir)
	node_dir = os.path.join(node_line_dir, "query_expansion")
	if not os.path.exists(node_dir):
		os.makedirs(node_dir)
	project_dir = pathlib.PurePath(node_line_dir).parent.parent

	# run query expansion
	results, execution_times = zip(
		*map(
			lambda task: measure_speed(
				task[0].run_evaluator,
				project_dir=project_dir,
				previous_result=previous_result,
				**task[1],
			),
			zip(modules, module_params),
		)
	)
	average_times = list(map(lambda x: x / len(results[0]), execution_times))

	# save results to folder
	pseudo_module_params = deepcopy(module_params)
	for i, module_param in enumerate(pseudo_module_params):
		if "prompt" in module_params:
			module_param["prompt"] = str(i)
	filepaths = list(
		map(lambda x: os.path.join(node_dir, f"{x}.parquet"), range(len(modules)))
	)
	list(
		map(lambda x: x[0].to_parquet(x[1], index=False), zip(results, filepaths))
	)  # execute save to parquet
	filenames = list(map(lambda x: os.path.basename(x), filepaths))

	# make summary file
	summary_df = pd.DataFrame(
		{
			"filename": filenames,
			"module_name": list(map(lambda module: module.__name__, modules)),
			"module_params": module_params,
			"execution_time": average_times,
		}
	)

	# Run evaluation when there are more than one module.
	if len(modules) > 1:
		# pop general keys from strategies (e.g. metrics, speed_threshold)
		general_key = ["metrics", "speed_threshold", "strategy"]
		general_strategy = dict(
			filter(lambda x: x[0] in general_key, strategies.items())
		)
		extra_strategy = dict(
			filter(lambda x: x[0] not in general_key, strategies.items())
		)

		# first, filter by threshold if it is enabled.
		if general_strategy.get("speed_threshold") is not None:
			results, filenames = filter_by_threshold(
				results, average_times, general_strategy["speed_threshold"], filenames
			)

		# check metrics in strategy
		if general_strategy.get("metrics") is None:
			raise ValueError(
				"You must at least one metrics for query expansion evaluation."
			)

		if extra_strategy.get("top_k") is None:
			extra_strategy["top_k"] = 10  # default value

		# get retrieval modules from strategy
		retrieval_callables, retrieval_params = make_retrieval_callable_params(
			extra_strategy
		)

		# get retrieval_gt
		retrieval_gt = pd.read_parquet(
			os.path.join(project_dir, "data", "qa.parquet"), engine="pyarrow"
		)["retrieval_gt"].tolist()

		# make rows to metric_inputs
		metric_inputs = [
			MetricInput(retrieval_gt=ret_gt, query=query, generation_gt=gen_gt)
			for ret_gt, query, gen_gt in zip(
				retrieval_gt,
				previous_result["query"].tolist(),
				previous_result["generation_gt"].tolist(),
			)
		]

		# run evaluation
		evaluation_results = list(
			map(
				lambda result: evaluate_one_query_expansion_node(
					retrieval_callables,
					retrieval_params,
					[
						setattr(metric_input, "queries", queries) or metric_input
						for metric_input, queries in zip(
							metric_inputs, result["queries"].to_list()
						)
					],
					general_strategy["metrics"],
					project_dir,
					previous_result,
					general_strategy.get("strategy", "mean"),
				),
				results,
			)
		)

		evaluation_df = pd.DataFrame(
			{
				"filename": filenames,
				**{
					f"query_expansion_{metric_name}": list(
						map(lambda x: x[metric_name].mean(), evaluation_results)
					)
					for metric_name in general_strategy["metrics"]
				},
			}
		)
		summary_df = pd.merge(
			on="filename", left=summary_df, right=evaluation_df, how="left"
		)

		best_result, best_filename = select_best(
			evaluation_results,
			general_strategy["metrics"],
			filenames,
			strategies.get("strategy", "mean"),
		)
		# change metric name columns to query_expansion_metric_name
		best_result = best_result.rename(
			columns={
				metric_name: f"query_expansion_{metric_name}"
				for metric_name in strategies["metrics"]
			}
		)
		best_result = best_result.drop(
			columns=["retrieved_contents", "retrieved_ids", "retrieve_scores"]
		)
	else:
		best_result, best_filename = results[0], filenames[0]
		best_result = pd.concat([previous_result, best_result], axis=1)

	# add 'is_best' column at summary file
	summary_df["is_best"] = summary_df["filename"] == best_filename

	# save files
	summary_df.to_csv(os.path.join(node_dir, "summary.csv"), index=False)
	best_result.to_parquet(
		os.path.join(node_dir, f"best_{os.path.splitext(best_filename)[0]}.parquet"),
		index=False,
	)

	return best_result


def evaluate_one_query_expansion_node(
	retrieval_funcs: List,
	retrieval_params: List[Dict],
	metric_inputs: List[MetricInput],
	metrics: List[str],
	project_dir,
	previous_result: pd.DataFrame,
	strategy_name: str,
) -> pd.DataFrame:
	previous_result["queries"] = [
		metric_input.queries for metric_input in metric_inputs
	]
	retrieval_results = list(
		map(
			lambda x: x[0].run_evaluator(
				project_dir=project_dir, previous_result=previous_result, **x[1]
			),
			zip(retrieval_funcs, retrieval_params),
		)
	)
	# Cast each retrieval results
	retrieval_result_dicts = list(map(cast_retrieve_infos, retrieval_results))
	evaluation_results = list(
		map(
			lambda x: evaluate_retrieval_node(
				x,
				metric_inputs,
				metrics,
			),
			retrieval_result_dicts,
		)
	)
	best_result, _ = select_best(
		evaluation_results, metrics, strategy_name=strategy_name
	)
	best_result = pd.concat([previous_result, best_result], axis=1)
	return best_result


def make_retrieval_callable_params(strategy_dict: Dict):
	"""
	strategy_dict looks like this:

	.. Code:: json

	    {
	        "metrics": ["retrieval_f1", "retrieval_recall"],
	        "top_k": 50,
	        "retrieval_modules": [
	          {"module_type": "bm25"},
	          {"module_type": "vectordb", "embedding_model": ["openai", "huggingface"]}
	        ]
	      }

	"""
	node_dict = deepcopy(strategy_dict)
	retrieval_module_list: Optional[List[Dict]] = node_dict.pop(
		"retrieval_modules", None
	)
	if retrieval_module_list is None:
		retrieval_module_list = [
			{
				"module_type": "bm25",
			}
		]
	node_params = node_dict
	modules = list(
		map(
			lambda module_dict: get_support_modules(module_dict.pop("module_type")),
			retrieval_module_list,
		)
	)
	param_combinations = list(
		map(
			lambda module_dict: make_combinations({**module_dict, **node_params}),
			retrieval_module_list,
		)
	)
	return explode(modules, param_combinations)


def evaluate_retrieval_node(
	result_dict: Dict,
	metric_inputs: List[MetricInput],
	metrics: Union[List[str], List[Dict]],
) -> pd.DataFrame:
	"""
	Evaluate retrieval node from retrieval node result dataframe.

	:param result_df: The result dataframe from a retrieval node.
	:param metric_inputs: List of metric input schema for AutoRAG.
	:param metrics: Metric list from input strategies.
	:return: Return result_df with metrics columns.
	    The columns will be 'retrieved_contents', 'retrieved_ids', 'retrieve_scores', and metric names.
	"""

	@evaluate_retrieval(
		metric_inputs=metric_inputs,
		metrics=metrics,
	)
	def evaluate_this_module(_dict: Dict):
		return (
			_dict["retrieved_contents"],
			_dict["retrieved_ids"],
			_dict["retrieve_scores"],
		)

	return evaluate_this_module(result_dict)



from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pandas as pd

from autorag.embedding.base import EmbeddingModel
from autorag.evaluation.metric.util import calculate_cosine_similarity
from autorag.nodes.passagefilter.base import BasePassageFilter
from autorag.nodes.passagefilter.similarity_threshold_cutoff import (
	embedding_query_content,
)
from autorag.utils import result_to_dataframe
from autorag.utils.util import empty_cuda_cache, pop_params


class SimilarityPercentileCutoff(BasePassageFilter):
	def __init__(self, project_dir: Union[str, Path], *args, **kwargs):
		"""
		Initialize the SimilarityPercentileCutoff module

		:param project_dir: The project directory to use for initializing the module
		:param embedding_model: The embedding model string to use for calculating similarity
		        Default is "openai" which is OpenAI text-embedding-ada-002 embedding model.
		"""
		super().__init__(project_dir, *args, **kwargs)
		embedding_model = kwargs.pop("embedding_model", "openai")
		self.embedding_model = EmbeddingModel.load(embedding_model)()

	def __del__(self):
		super().__del__()
		del self.embedding_model

		empty_cuda_cache()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, **kwargs):
		queries, contents, scores, ids = self.cast_to_run(previous_result)
		kwargs = pop_params(self._pure, kwargs)
		return self._pure(queries, contents, scores, ids, **kwargs)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		scores_list: List[List[float]],
		ids_list: List[List[str]],
		percentile: float,
		batch: int = 128,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Re-calculate each content's similarity with the query and filter out the contents that are below the content's
		length times percentile. If This is a filter and does not override scores. The output of scores is not coming from
		query-content similarity.
		If the value of content's length times percentile is less than 1, keep the only one highest similarity content.

		:param queries: The list of queries to use for filtering
		:param contents_list: The list of lists of contents to filter
		:param scores_list: The list of lists of scores retrieved
		:param ids_list: The list of lists of ids retrieved
		:param percentile: The percentile to cut off
		:param batch: The number of queries to be processed in a batch
		    Default is 128.
		:return: Tuple of lists containing the filtered contents, ids, and scores
		"""
		query_embeddings, content_embeddings = embedding_query_content(
			queries, contents_list, self.embedding_model, batch
		)

		results = list(
			map(
				lambda x: self.__row_pure(x[0], x[1], x[2], x[3], x[4], percentile),
				zip(
					query_embeddings,
					content_embeddings,
					contents_list,
					ids_list,
					scores_list,
				),
			)
		)

		remain_content_list = list(map(lambda x: x[0], results))
		remain_ids_list = list(map(lambda x: x[1], results))
		remain_scores_list = list(map(lambda x: x[2], results))

		return remain_content_list, remain_ids_list, remain_scores_list

	@staticmethod
	def __row_pure(
		query_embedding: str,
		content_embeddings: List[List[float]],
		content_list: List[str],
		ids_list: List[str],
		scores_list: List[float],
		percentile: float,
	) -> Tuple[List[str], List[str], List[float]]:
		"""
		Return tuple of lists containing the filtered contents, ids, and scores

		:param query_embedding: Query embedding
		:param content_embeddings: Each content embedding
		:param content_list: Each content
		:param ids_list: Each id
		:param scores_list: Each score
		:param percentile: The percentile to cut off
		:return: Tuple of lists containing the filtered contents, ids, and scores
		"""
		num_top_k = int(len(content_embeddings) * percentile)

		if num_top_k == 0:
			num_top_k = 1

		similarities = np.array(
			list(
				map(
					lambda x: calculate_cosine_similarity(query_embedding, x),
					content_embeddings,
				)
			)
		).tolist()

		content_id_score_similarity = list(
			zip(ids_list, content_list, scores_list, similarities)
		)

		sorted_content_id_score_similarity = sorted(
			content_id_score_similarity, key=lambda x: x[3], reverse=True
		)[:num_top_k]

		content_result, id_result, score_result, _ = zip(
			*sorted_content_id_score_similarity
		)
		return list(content_result), list(id_result), list(score_result)



from copy import deepcopy
from dataclasses import dataclass, field
from typing import Callable, Dict

from autorag.support import get_support_modules


@dataclass
class Module:
	module_type: str
	module_param: Dict
	module: Callable = field(init=False)

	def __post_init__(self):
		self.module = get_support_modules(self.module_type)
		if self.module is None:
			raise ValueError(f"Module type {self.module_type} is not supported.")

	@classmethod
	def from_dict(cls, module_dict: Dict) -> "Module":
		_module_dict = deepcopy(module_dict)
		module_type = _module_dict.pop("module_type")
		module_params = _module_dict
		return cls(module_type, module_params)



import logging
import os
import pathlib
from typing import List, Dict

import pandas as pd

from autorag.nodes.retrieval.run_util import evaluate_retrieval_node
from autorag.schema.metricinput import MetricInput
from autorag.strategy import measure_speed, filter_by_threshold, select_best
from autorag.utils.util import apply_recursive, to_list

logger = logging.getLogger("AutoRAG")


def run_passage_reranker_node(
	modules: List,
	module_params: List[Dict],
	previous_result: pd.DataFrame,
	node_line_dir: str,
	strategies: Dict,
) -> pd.DataFrame:
	"""
	Run evaluation and select the best module among passage reranker node results.

	:param modules: Passage reranker modules to run.
	:param module_params: Passage reranker module parameters.
	:param previous_result: Previous result dataframe.
	    Could be retrieval, reranker modules result.
	    It means it must contain 'query', 'retrieved_contents', 'retrieved_ids', 'retrieve_scores' columns.
	:param node_line_dir: This node line's directory.
	:param strategies: Strategies for passage reranker node.
	    In this node, we use 'retrieval_f1', 'retrieval_recall' and 'retrieval_precision'.
	    You can skip evaluation when you use only one module and a module parameter.
	:return: The best result dataframe with previous result columns.
	"""
	if not os.path.exists(node_line_dir):
		os.makedirs(node_line_dir)
	project_dir = pathlib.PurePath(node_line_dir).parent.parent
	qa_df = pd.read_parquet(
		os.path.join(project_dir, "data", "qa.parquet"), engine="pyarrow"
	)
	retrieval_gt = qa_df["retrieval_gt"].tolist()
	retrieval_gt = apply_recursive(lambda x: str(x), to_list(retrieval_gt))

	# make rows to metric_inputs
	metric_inputs = [
		MetricInput(retrieval_gt=ret_gt, query=query, generation_gt=gen_gt)
		for ret_gt, query, gen_gt in zip(
			retrieval_gt, qa_df["query"].tolist(), qa_df["generation_gt"].tolist()
		)
	]

	results, execution_times = zip(
		*map(
			lambda task: measure_speed(
				task[0].run_evaluator,
				project_dir=project_dir,
				previous_result=previous_result,
				**task[1],
			),
			zip(modules, module_params),
		)
	)
	average_times = list(map(lambda x: x / len(results[0]), execution_times))

	# run metrics before filtering
	if strategies.get("metrics") is None:
		raise ValueError(
			"You must at least one metrics for passage_reranker evaluation."
		)
	results = list(
		map(
			lambda x: evaluate_retrieval_node(
				x,
				metric_inputs,
				strategies.get("metrics"),
			),
			results,
		)
	)

	# save results to folder
	save_dir = os.path.join(node_line_dir, "passage_reranker")  # node name
	if not os.path.exists(save_dir):
		os.makedirs(save_dir)
	filepaths = list(
		map(lambda x: os.path.join(save_dir, f"{x}.parquet"), range(len(modules)))
	)
	list(
		map(lambda x: x[0].to_parquet(x[1], index=False), zip(results, filepaths))
	)  # execute save to parquet
	filenames = list(map(lambda x: os.path.basename(x), filepaths))

	summary_df = pd.DataFrame(
		{
			"filename": filenames,
			"module_name": list(map(lambda module: module.__name__, modules)),
			"module_params": module_params,
			"execution_time": average_times,
			**{
				f"passage_reranker_{metric}": list(
					map(lambda result: result[metric].mean(), results)
				)
				for metric in strategies.get("metrics")
			},
		}
	)

	# filter by strategies
	if strategies.get("speed_threshold") is not None:
		results, filenames = filter_by_threshold(
			results, average_times, strategies["speed_threshold"], filenames
		)
	selected_result, selected_filename = select_best(
		results,
		strategies.get("metrics"),
		filenames,
		strategies.get("strategy", "mean"),
	)
	# change metric name columns to passage_reranker_metric_name
	selected_result = selected_result.rename(
		columns={
			metric_name: f"passage_reranker_{metric_name}"
			for metric_name in strategies["metrics"]
		}
	)
	# drop retrieval result columns in previous_result
	previous_result = previous_result.drop(
		columns=["retrieved_contents", "retrieved_ids", "retrieve_scores"]
	)
	best_result = pd.concat([previous_result, selected_result], axis=1)

	# add 'is_best' column to summary file
	summary_df["is_best"] = summary_df["filename"] == selected_filename

	# save files
	summary_df.to_csv(os.path.join(save_dir, "summary.csv"), index=False)
	best_result.to_parquet(
		os.path.join(
			save_dir, f"best_{os.path.splitext(selected_filename)[0]}.parquet"
		),
		index=False,
	)
	return best_result



import logging
from typing import List, Tuple

import pandas as pd

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils import result_to_dataframe
from autorag.utils.util import select_top_k, sort_by_scores, empty_cuda_cache

logger = logging.getLogger("AutoRAG")


class Upr(BasePassageReranker):
	def __init__(
		self,
		project_dir: str,
		use_bf16: bool = False,
		prefix_prompt: str = "Passage: ",
		suffix_prompt: str = "Please write a question based on this passage.",
		*args,
		**kwargs,
	):
		"""
		Initialize the UPR reranker node.

		:param project_dir: The project directory
		:param use_bf16: Whether to use bfloat16 for the model. Default is False.
		:param prefix_prompt: The prefix prompt for the language model that generates question for reranking.
			Default is "Passage: ".
			The prefix prompt serves as the initial context or instruction for the language model.
			It sets the stage for what is expected in the output
		:param suffix_prompt: The suffix prompt for the language model that generates question for reranking.
			Default is "Please write a question based on this passage.".
			The suffix prompt provides a cue or a closing instruction to the language model,
				signaling how to conclude the generated text or what format to follow at the end.
		:param kwargs: Extra arguments
		"""
		super().__init__(project_dir, *args, **kwargs)

		self.scorer = UPRScorer(
			suffix_prompt=suffix_prompt, prefix_prompt=prefix_prompt, use_bf16=use_bf16
		)

	def __del__(self):
		del self.scorer
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		return self._pure(queries, contents, ids, top_k)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents based on their relevance to a query using UPR.
		UPR is a reranker based on UPR (https://github.com/DevSinghSachan/unsupervised-passage-reranking).
		The language model will make a question based on the passage and rerank the passages by the likelihood of the question.
		The default model is t5-large.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved

		:return: tuple of lists containing the reranked contents, ids, and scores
		"""
		df = pd.DataFrame(
			{
				"query": queries,
				"contents": contents_list,
				"ids": ids_list,
			}
		)

		df["scores"] = df.apply(
			lambda row: self.scorer.compute(
				query=row["query"], contents=row["contents"]
			),
			axis=1,
		)
		df[["contents", "ids", "scores"]] = df.apply(
			lambda x: sort_by_scores(x, reverse=False), axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)
		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


class UPRScorer:
	def __init__(self, suffix_prompt: str, prefix_prompt: str, use_bf16: bool = False):
		try:
			import torch
			from transformers import T5Tokenizer, T5ForConditionalGeneration
		except ImportError:
			raise ImportError(
				"torch is not installed. Please install torch to use UPRReranker."
			)
		model_name = "t5-large"
		self.device = "cuda" if torch.cuda.is_available() else "cpu"
		self.tokenizer = T5Tokenizer.from_pretrained(model_name)
		self.model = T5ForConditionalGeneration.from_pretrained(
			model_name, torch_dtype=torch.bfloat16 if use_bf16 else torch.float32
		).to(self.device)
		self.suffix_prompt = suffix_prompt
		self.prefix_prompt = prefix_prompt

	def compute(self, query: str, contents: List[str]) -> List[float]:
		try:
			import torch
		except ImportError:
			raise ImportError(
				"torch is not installed. Please install torch to use UPRReranker."
			)
		query_token = self.tokenizer(
			query, max_length=128, truncation=True, return_tensors="pt"
		)
		prompts = list(
			map(
				lambda content: f"{self.prefix_prompt} {content} {self.suffix_prompt}",
				contents,
			)
		)
		prompt_token_outputs = self.tokenizer(
			prompts,
			padding="longest",
			max_length=512,
			pad_to_multiple_of=8,
			truncation=True,
			return_tensors="pt",
		)

		query_input_ids = torch.repeat_interleave(
			query_token["input_ids"], len(contents), dim=0
		).to(self.device)

		with torch.no_grad():
			logits = self.model(
				input_ids=prompt_token_outputs["input_ids"].to(self.device),
				attention_mask=prompt_token_outputs["attention_mask"].to(self.device),
				labels=query_input_ids,
			).logits
		log_softmax = torch.nn.functional.log_softmax(logits, dim=-1)
		nll = -log_softmax.gather(2, query_input_ids.unsqueeze(2)).squeeze(2)
		avg_nll = torch.sum(nll, dim=1)
		return avg_nll.tolist()

	def __del__(self):
		del self.model
		del self.tokenizer
		empty_cuda_cache()



import logging
import sys

from random import random
from typing import List, Union, Dict

from llama_index.core.embeddings.mock_embed_model import MockEmbedding
from llama_index.embeddings.openai import OpenAIEmbedding
from llama_index.embeddings.openai import OpenAIEmbeddingModelType
from llama_index.embeddings.ollama import OllamaEmbedding
from langchain_openai.embeddings import OpenAIEmbeddings
from llama_index.embeddings.openai_like import OpenAILikeEmbedding

from autorag import LazyInit

logger = logging.getLogger("AutoRAG")


class MockEmbeddingRandom(MockEmbedding):
	"""Mock embedding with random vectors."""

	def _get_vector(self) -> List[float]:
		return [random() for _ in range(self.embed_dim)]


embedding_models = {
	# llama index
	"openai": LazyInit(
		OpenAIEmbedding
	),  # default model is OpenAIEmbeddingModelType.TEXT_EMBED_ADA_002
	"openai_embed_3_large": LazyInit(
		OpenAIEmbedding, model_name=OpenAIEmbeddingModelType.TEXT_EMBED_3_LARGE
	),
	"openai_embed_3_small": LazyInit(
		OpenAIEmbedding, model_name=OpenAIEmbeddingModelType.TEXT_EMBED_3_SMALL
	),
	"mock": LazyInit(MockEmbeddingRandom, embed_dim=768),
	# langchain
	"openai_langchain": LazyInit(OpenAIEmbeddings),
	"ollama": LazyInit(OllamaEmbedding),
	# openai like
	"openai_like": LazyInit(OpenAILikeEmbedding),
}

try:
	# you can use your own model in this way.
	from llama_index.embeddings.huggingface import HuggingFaceEmbedding

	embedding_models["huggingface_baai_bge_small"] = LazyInit(
		HuggingFaceEmbedding, model_name="BAAI/bge-small-en-v1.5"
	)
	embedding_models["huggingface_cointegrated_rubert_tiny2"] = LazyInit(
		HuggingFaceEmbedding, model_name="cointegrated/rubert-tiny2"
	)
	embedding_models["huggingface_all_mpnet_base_v2"] = LazyInit(
		HuggingFaceEmbedding,
		model_name="sentence-transformers/all-mpnet-base-v2",
		max_length=512,
	)
	embedding_models["huggingface_bge_m3"] = LazyInit(
		HuggingFaceEmbedding, model_name="BAAI/bge-m3"
	)
	embedding_models["huggingface_multilingual_e5_large"] = LazyInit(
		HuggingFaceEmbedding, model_name="intfloat/multilingual-e5-large-instruct"
	)
except ImportError:
	logger.info(
		"You are using API version of AutoRAG."
		"To use local version, run pip install 'AutoRAG[gpu]'"
	)


class EmbeddingModel:
	@staticmethod
	def load(config: Union[str, Dict, List[Dict]]):
		if isinstance(config, str):
			return EmbeddingModel.load_from_str(config)
		elif isinstance(config, dict):
			return EmbeddingModel.load_from_dict(config)
		elif isinstance(config, list):
			return EmbeddingModel.load_from_list(config)
		else:
			raise ValueError("Invalid type of config")

	@staticmethod
	def load_from_str(name: str):
		try:
			return embedding_models[name]
		except KeyError:
			raise ValueError(f"Embedding model '{name}' is not supported")

	@staticmethod
	def load_from_list(option: List[dict]):
		if len(option) != 1:
			raise ValueError("Only one embedding model is supported")
		return EmbeddingModel.load_from_dict(option[0])

	@staticmethod
	def load_from_dict(option: dict):
		def _check_keys(target: dict):
			if "type" not in target or "model_name" not in target:
				raise ValueError("Both 'type' and 'model_name' must be provided")
			if target["type"] not in ["openai", "huggingface", "mock", "ollama"]:
				raise ValueError(
					f"Embedding model type '{target['type']}' is not supported"
				)

		def _get_huggingface_class():
			module = sys.modules.get("llama_index.embeddings.huggingface")
			if not module:
				logger.info(
					"You are using API version of AutoRAG. "
					"To use local version, run `pip install 'AutoRAG[gpu]'`."
				)
				return None
			return getattr(module, "HuggingFaceEmbedding", None)

		_check_keys(option)

		model_options = option
		model_type = model_options.pop("type")

		embedding_map = {
			"openai": OpenAIEmbedding,
			"mock": MockEmbeddingRandom,
			"huggingface": _get_huggingface_class(),
			"ollama": OllamaEmbedding,
			"openai_like": OpenAILikeEmbedding,
		}

		embedding_class = embedding_map.get(model_type)
		if not embedding_class:
			raise ValueError(f"Embedding model type '{model_type}' is not supported")

		return LazyInit(embedding_class, **model_options)



from typing import List, Tuple

import pandas as pd

from autorag.nodes.passagefilter.base import BasePassageFilter
from autorag.utils.util import sort_by_scores, select_top_k, result_to_dataframe


class PercentileCutoff(BasePassageFilter):
	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, scores, ids = self.cast_to_run(previous_result)
		return self._pure(queries, contents, scores, ids, *args, **kwargs)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		scores_list: List[List[float]],
		ids_list: List[List[str]],
		percentile: float,
		reverse: bool = False,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Filter out the contents that are below the content's length times percentile.
		If This is a filter and does not override scores.
		If the value of content's length times percentile is less than 1, keep the only one highest similarity content.

		:param queries: The list of queries to use for filtering
		:param contents_list: The list of lists of contents to filter
		:param scores_list: The list of lists of scores retrieved
		:param ids_list: The list of lists of ids retrieved
		:param percentile: The percentile to cut off
		:param reverse: If True, the lower the score, the better
		    Default is False.
		:return: Tuple of lists containing the filtered contents, ids, and scores
		"""
		num_top_k = max(1, int(len(scores_list[0]) * percentile))

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": scores_list,
			}
		)

		reverse = not reverse
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand", reverse=reverse
		)
		results = select_top_k(df, ["contents", "ids", "scores"], num_top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)



import uuid
from typing import Iterable

import pandas as pd


def random_single_hop(
	corpus_df: pd.DataFrame, n: int, random_state: int = 42
) -> pd.DataFrame:
	sample_df = corpus_df.sample(n, random_state=random_state)
	return pd.DataFrame(
		{
			"qid": [str(uuid.uuid4()) for _ in range(len(sample_df))],
			"retrieval_gt": [[[id_]] for id_ in sample_df["doc_id"].tolist()],
		}
	)


def range_single_hop(corpus_df: pd.DataFrame, idx_range: Iterable):
	sample_df = corpus_df.iloc[idx_range]
	return pd.DataFrame(
		{
			"qid": [str(uuid.uuid4()) for _ in range(len(sample_df))],
			"retrieval_gt": [[[id_]] for id_ in sample_df["doc_id"].tolist()],
		}
	)



import os
import pathlib
import uuid
from typing import Callable

import pandas as pd


def generate_qa_row(llm, corpus_data_row):
	"""
	this sample code to generate rag dataset using OpenAI chat model

	:param llm: guidance model
	:param corpus_data_row: need "contents" column
	:return: should to be dict which has "query", "generation_gt" columns at least.
	"""
	from guidance import gen
	import guidance

	temp_llm = llm
	with guidance.user():
		temp_llm += f"""
    You have to found a passge to solve "the problem".
    You need to build a clean and clear set of (problem, passage, answer) in json format
    so that you don't have to ask about "the problem" again.
    problem need to end with question mark("?").
    The process of approaching the answer based on the information of the given passage
    must be clearly and neatly displayed in the answer.\n
    \n
    Here is set of (problem, passage, answer) in JSON format:\n
    {{\n
        "passage": {corpus_data_row["contents"]}\n
        "problem":
    """

	with guidance.assistant():
		temp_llm += gen("query", stop="?")
	with guidance.user():
		temp_llm += """
        "answer":
        """
	with guidance.assistant():
		temp_llm += gen("generation_gt")

	corpus_data_row["metadata"]["qa_generation"] = "simple"

	response = {"query": temp_llm["query"], "generation_gt": temp_llm["generation_gt"]}
	return response


def generate_simple_qa_dataset(
	llm,
	corpus_data: pd.DataFrame,
	output_filepath: str,
	generate_row_function: Callable,
	**kwargs,
):
	"""
	corpus_data to qa_dataset
	qa_dataset will be saved to filepath(file_dir/filename)

	:param llm: guidance.models.Model
	:param corpus_data: pd.DataFrame. refer to the basic structure
	:param output_filepath: file_dir must exist, filepath must not exist. file extension must be .parquet
	:param generate_row_function: input(llm, corpus_data_row, kwargs) output(dict[columns contain "query" and "generation_gt"])
	:param kwargs: if generate_row_function requires more args, use kwargs
	:return: qa_dataset as pd.DataFrame
	"""
	output_file_dir = pathlib.PurePath(output_filepath).parent
	if not os.path.isdir(output_file_dir):
		raise NotADirectoryError(f"directory {output_file_dir}  not found.")
	if not output_filepath.endswith("parquet"):
		raise NameError(
			f'file path: {output_filepath}  filename extension need to be ".parquet"'
		)
	if os.path.exists(output_filepath):
		raise FileExistsError(
			f"{output_filepath.split('/')[-1]} already exists in {output_file_dir}."
		)

	qa_data_lst = []
	for _, corpus_data_row in corpus_data.iterrows():
		response = generate_row_function(
			llm=llm, corpus_data_row=corpus_data_row, **kwargs
		)
		qa_data_lst.append(
			{
				"qid": str(uuid.uuid4()),
				"query": response["query"],
				"retrieval_gt": [[corpus_data_row["doc_id"]]],
				"generation_gt": [response["generation_gt"]],
				"metadata": corpus_data_row["metadata"],
			}
		)

	qa_dataset = pd.DataFrame(qa_data_lst)
	qa_dataset.to_parquet(output_filepath, index=False)

	return qa_dataset



import uuid
from typing import Optional

import pandas as pd
from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseChatModel
from langchain_openai import ChatOpenAI, OpenAIEmbeddings

from autorag.data.utils.util import corpus_df_to_langchain_documents
from autorag.utils import cast_qa_dataset


def generate_qa_ragas(
	corpus_df: pd.DataFrame,
	test_size: int,
	distributions: Optional[dict] = None,
	generator_llm: Optional[BaseChatModel] = None,
	critic_llm: Optional[BaseChatModel] = None,
	embedding_model: Optional[Embeddings] = None,
	**kwargs,
) -> pd.DataFrame:
	"""
	QA dataset generation using RAGAS.
	Returns qa dataset dataframe.

	:param corpus_df: Corpus dataframe.
	:param test_size: Number of queries to generate.
	:param distributions: Distributions of different types of questions.
	    Default is "simple is 0.5, multi_context is 0.4, and reasoning is 0.1."
	    Each type of questions refers to Ragas evolution types.
	:param generator_llm: Generator language model from Langchain.
	:param critic_llm: Critic language model from Langchain.
	:param embedding_model: Embedding model from Langchain.
	:param kwargs: The additional option to pass to the 'generate_with_langchain_docs' method.
	    You can input 'with_debugging_logs', 'is_async', 'raise_exceptions', and 'run_config'.
	:return: QA dataset dataframe.
	"""
	from ragas.testset import TestsetGenerator
	from ragas.testset.evolutions import simple, reasoning, multi_context

	if generator_llm is None:
		generator_llm = ChatOpenAI(model="gpt-3.5-turbo-16k")
	if critic_llm is None:
		critic_llm = ChatOpenAI(model="gpt-4-turbo")
	if embedding_model is None:
		embedding_model = OpenAIEmbeddings()
	if distributions is None:
		distributions = {simple: 0.5, multi_context: 0.4, reasoning: 0.1}

	assert sum(list(distributions.values())) == 1.0, "Sum of distributions must be 1.0"

	generator = TestsetGenerator.from_langchain(
		generator_llm, critic_llm, embedding_model
	)

	langchain_docs = corpus_df_to_langchain_documents(corpus_df)

	test_df = generator.generate_with_langchain_docs(
		langchain_docs, test_size, distributions=distributions, **kwargs
	).to_pandas()

	result_df = pd.DataFrame(
		{
			"qid": [str(uuid.uuid4()) for _ in range(len(test_df))],
			"query": test_df["question"].tolist(),
			"generation_gt": list(map(lambda x: x, test_df["ground_truth"].tolist())),
		}
	)

	result_df["retrieval_gt"] = test_df["metadata"].apply(
		lambda x: list(map(lambda y: y["filename"], x))
	)
	result_df = cast_qa_dataset(result_df)

	return result_df



import copy
from typing import List

import pandas as pd

from autorag.nodes.promptmaker.base import BasePromptMaker
from autorag.utils import result_to_dataframe


class ChatFstring(BasePromptMaker):
	@result_to_dataframe(["prompts"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		query, retrieved_contents, prompt = self.cast_to_run(
			previous_result, *args, **kwargs
		)
		return self._pure(prompt, query, retrieved_contents)

	def _pure(
		self,
		prompt: List[dict[str, str]],
		queries: List[str],
		retrieved_contents: List[List[str]],
	) -> List[List[dict[str, str]]]:
		"""
		Make a prompt using f-string from a query and retrieved_contents.
		You must type a prompt or prompt list at a config YAML file like this:

		.. Code:: yaml
		nodes:
		- node_type: prompt_maker
		  modules:
		  - module_type: chatfstring
			prompt:
			- - role: system
			  content: You are a helpful assistant that helps people find information.
			  - role: user
			    content: |
			    Answer this question: {query}
			    {retrieved_contents}
			- - role: system
			    content: You are helpful.
			  - role: user
				content: |
			    Read the passages carefully and answer this question: {query}

			    Passages: {retrieved_contents}

		:param prompt: A prompt string.
		:param queries: List of query strings.
		:param retrieved_contents: List of retrieved contents.
		:return: Prompts that are made by chat f-string.
			It is the list of OpenAI chat format prompts.
		"""

		def fstring_row(
			_prompt: List[dict[str, str]], _query: str, _retrieved_contents: List[str]
		) -> List[dict[str, str]]:
			contents_str = "\n\n".join(_retrieved_contents)
			result_prompt = copy.deepcopy(_prompt)
			for lst in result_prompt:
				if "content" in lst:
					lst["content"] = lst["content"].format(
						query=_query, retrieved_contents=contents_str
					)

			return result_prompt

		return list(
			map(
				lambda x: fstring_row(prompt, x[0], x[1]),
				zip(queries, retrieved_contents),
			)
		)
